/*  test_matrix_mult.c  (Mon Aug 30 15:30:51 EDT 1999)
 *  Author:  Galen Wilkerson
 *
 *  See README.txt for usage
 *
 *  Significant parts of this code modified from example_read.c
 *  by Roldan Pozo and Karin Remington
 *
 *  Contact:  galen@cam.nist.gov or pozo@cam.nist.gov
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "mmio.h"
#include "spblas_i.h"
#include "c_blas.h"

/* initialize a matrix */
void ArraySetToValue(double *A, int ldA, int outer , int inner, double value)
{
  double *p, *pend, *t, *tend;

  pend = A + outer*ldA;
  for (p=A; p<pend; p+=ldA)  {
    tend = p+inner;
    for (t=p; t<tend; *t++ = value);
  }
}

/* print a dense matrix */
void ArrayPrint(double *A, int ldA, int outer , int inner, FILE *output)
{
  double *p, *pend, *t, *tend;
  
  fprintf(output, "%d %d\n", outer, inner);

  pend = A + outer*ldA;

  for (p=A; p<pend; p+=ldA) {
    tend = p+inner;
    for (t=p; t<tend; t++) 
      fprintf(output,"%20.15g ",*t);
    fprintf(output,"\n");
  }
}

/* save a dense matrix */
void ArraySave(double *A, int ldA, int numRows , int numCols, char
	       *outFileName, enum blas_order_type order)
{
  double *p, *pend, *t, *tend;
  FILE *outFile;

  if ((outFile = fopen(outFileName, "w")) == NULL) 
    exit(1);

  fprintf(outFile,"%d %d\n", numRows, numCols);
  
  if (order == blas_colmajor) {
    pend = A + numCols*ldA;
    for (p=A; p<pend; p+=ldA) {
      tend = p + numRows;
      for (t=p; t<tend; t++) 
	fprintf(outFile,"%20.15g ",*t);
      fprintf(outFile,"\n");
    }
  }
  else if (order == blas_rowmajor) {
        
    tend = A + numCols;
    for (t = A; t< tend; t++) {
      pend = t + numRows * ldA;
      for (p = t; p < pend; p += ldA)
	fprintf(outFile,"%20.15g ",*p);
      fprintf(outFile,"\n");
    }
  }
  else 
    exit(1);

  fclose(outFile);  
}

/* read in a dense matrix */
void ArrayRead(double *A, int M , int N, FILE *input)
{
  int i;
  int end = M * N;
  
  fscanf(input, "%d %d\n", &M, &N);
  for (i = 0; i < M*N; i++) {
    fscanf(input, "%lg", &A[i]);
  }
}


DCSC DCOO_2_DCSC(DCOO A);


int main(int argc, char *argv[])
{
    int ret_code;
    MM_typecode matcode;

    FILE *f, *inB, *inBB, *inBB2, *outC0, *outC1, *outC2, *outC3, *outC4, *outC5,
      *outC8, *outC9, *outC12, *outC13;

    FILE *outB12, *outB13;


    /* values to create the compressed sparse row matrix A */
    
    SPBLASI_double_CSC CSCA, CSCA1;

    /* values to create the compressed coordinate matrix A and A1*/
    int M, N, nz;   
    int i, j, temp, *I, *J, *I1, *J1;
    double *val;
    int properties = sparse_blas_zero_base;
    int transpose = blas_no_trans;
    int order = blas_colmajor;
    SPBLASI_double_COO A, A1;

    double alpha = 1.0;
    double *B, *BB, *BB2;
    int ldb = 30;
    int BBN = 60;
    int BBk = 60;
    int BB2N = 60;
    int BB2k = 60;

    /* k == the number of columns of B and C */
    int k = 30;
    
    double beta = 0.0;
    double *C, *CC;
    int ldc = 30;
    int CCM = 60;
    int CCk = 60;

    if ((argc > 1) && (strcmp(argv[1],"-h") == 0))
      {
	fprintf(stderr, "Requires subdirectory Data/ \nContaining the following files: \n B30rowmaj\n BB60random\n BB60rowmajor\n pores_1.mtx\n\n", argv[0]);
	
	exit(1);
      }
    
    
    if ((f = fopen("Data/pores_1.mtx", "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }
    if ((inB = fopen("Data/B30random", "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }
    if (mm_read_banner(f, &matcode) != 0)
    {
      printf("Could not process Matrix Market banner.\n");
      exit(1);
    }


    /*  This is how one can screen matrix types if their application */
    /*  only supports a subset of the Matrix Market data types.      */

    if (mm_is_complex(matcode) && mm_is_matrix(matcode) && 
            mm_is_sparse(matcode) )
    {
      printf("Sorry, this application does not support ");
      printf("Market Market type: [%s]\n", mm_typecode_to_str(matcode));
      exit(1);
    }

    /* find out size of sparse matrix .... */

    if ((ret_code = mm_read_mtx_crd_size(f, &M, &N, &nz)) !=0)
        exit(1);


    /* reserve memory for matrix */

    I = (int *) malloc(nz * sizeof(int));
    J = (int *) malloc(nz * sizeof(int));
    val = (double *) malloc(nz * sizeof(double));


    /* reserve memory for 1-base indices */
    I1 = (int *) malloc(nz * sizeof(int));
    J1 = (int *) malloc(nz * sizeof(int));

    /* NOTE: when reading in doubles, ANSI C requires the use of the "l"  */
    /*   specifier as in "%lg", "%lf", "%le", otherwise errors will occur */
    /*  (ANSI C X3.159-1989, Sec. 4.9.6.2, p. 136 lines 13-15)            */

    /* read the compressed coordinate matrix */
    for (i=0; i<nz; i++)
    {
        fscanf(f, "%d %d %lg\n", &I[i], &J[i], &val[i]);
        /* adjust from 1-based to 0-based */
	
	I1[i] = I[i];
	J1[i] = J[i];

	I[i]--;  
	J[i]--;
    }

    if (f !=stdin) fclose(f);

    /************************/
    /* now write out matrix */
    /************************/

    mm_write_banner(stdout, matcode);
    mm_write_mtx_crd_size(stdout, M, N, nz);
    for (i=0; i<nz; i++)
      fprintf(stdout, "%d %d %20.19g\n", I[i]+1, J[i]+1, val[i]);
    

    /************************************
      create the structure A (zero-based)
      ***********************************/

    A = create_DCOO(M, N, nz, val, I, J, properties);


    /**************************
      convert the DCOO structure 
      A into the DCSC structure CSCA
      
      *************************/  
    
    CSCA = DCOO_2_DCSC(A);
    free(A);

    /************************************
      create the structure A1 (one-based - case 1)
      ***********************************/

    properties = sparse_blas_one_base;
    A1 = create_DCOO(M, N, nz, val, I1, J1, properties);
    
    /**************************
      convert the DCOO structure 
      A into the DCSC structure CSCA
      
      *************************/  

    CSCA1 = DCOO_2_DCSC(A1);
    free(A1);
    /***********************
      allocate memory for B
     **********************/
    
    B = (double *) malloc(N * k * sizeof(double));
    
    /*************************************************
      read (and print) the matrix B in col-major order
      ************************************************/
    
    ArrayRead(B, N, k, inB);

    /***********************************
      write B to file in row-major order
      **********************************/

    ArraySave(B,ldb,N,k,"Data/B30rowmaj",blas_rowmajor);

    /*************
      allocate C
      **************/
    
    C = (double *) malloc(M * k * sizeof(double));

    /*******************************************************
      create and initialize BB, 
      which contains B as a submatrix (column-major - case 12)
      *******************************************************/

    BBN = 2 * N;
    BBk = 2 * k;
    
    BB = (double *) malloc(BBN * BBk * sizeof(double));
    
    if ((inBB = fopen("Data/BB60random", "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }
    /* read the matrix BB in col-major order */

    ArrayRead(BB, BBN , BBk, inBB);
    
    /* create CC, which contains the submatrix C */
    
    CCM = 2 * M;
    CCk = 2 * k;
    
    CC = (double *) malloc(CCM * CCk * sizeof(double));

    /************************************************
      create and initialize BB2, 
      which contains B as a submatrix (row-major - case 13)
      ***********************************************/
    
    BB2N = 2 * N;
    BB2k = 2 * k;
    
    BB2 = (double *) malloc(BB2N * BB2k * sizeof(double));
    
    if ((inBB2 = fopen("Data/BB60rowmaj", "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }

    /*************************************************
      read & print the matrix BB2 in col-major order
      ************************************************/
    
    ArrayRead(BB2, BB2N , BB2k, inBB2);


    /****************************************************************
      Case 0:  alpha == 1.0, beta = 0.0, zero base, no transpose, column-major
      
      initialize C
      call c_dusmm_COO()
      open the outC1 file
      save C to outC1
      ****************************************************************/

    /*********************************************
      initialize C in column-major order
      ********************************************/
    
    ArraySetToValue(C,ldc,M,k,0.0);
    
    /********************************************
      multiply the matrix A by B

      C <- A * B
      ********************************************/
    
    printf("\nCSCA:\n");
    DCSC_print(CSCA);
    

    printf("\nB:\n");
    ArrayPrint(B, 30, k, N, stdout);

    c_dusmm_CSC(order, transpose, k, alpha, CSCA, B, ldb, beta, C, ldc);

    printf("\nC:\n");
    ArrayPrint(C, 30, k, M, stdout);


    /********************
      save C to file
      *********************/

    ArraySave(C,ldc,M,k,"Data/outC0",order);

    /*************
      End Case 0
      ************/

 
    /****************************************************************
      
      Case 1:  base one

    *****************************************************************/

    ArrayRead(B, N , k, inB);

    ArraySetToValue(C,ldc,M,k,0.0);
    
    c_dusmm_CSC(order, transpose, k, alpha, CSCA1, B, ldb, beta, C, ldc);

    ArraySave(C,ldc,M,k,"Data/outC1",order);

    /************
      END CASE 1
    *************/


    /****************************************************************
      
      Case 2:  alpha == 0.0 && beta != 0.0

      initialize C to non-zero value

      ****************************************************************/
    
    properties = sparse_blas_zero_base;
    alpha = 0.0;
    beta = -3.75;

    ArraySetToValue(C,ldc,M,k,1.0);
      
    c_dusmm_CSC(order, transpose, k, alpha, CSCA, B, ldb, beta, C, ldc);

    ArraySave(C,ldc,M,k,"Data/outC2",order);
    
    /***********
      END CASE 2
      ************/


    /************************************************************
     Case 3:	transpose
     
     *************************************************************/
    
    transpose = blas_trans;
    alpha = 1.0;
    beta = 0.0;

    ArraySetToValue(C,ldc,M,k,0.0);

    c_dusmm_CSC(order, transpose, k, alpha, CSCA, B, ldb, beta, C, ldc);

    ArraySave(C,ldc,M,k,"Data/outC3",order);

    /************
      END CASE 3
      ************/


    /***********************************************************
    Case 4 & Case 6:	ColMajor && beta == 0.0 && alpha == 1.0
      
    *************************************************************/

    transpose = blas_no_trans;
    alpha = 1.0;
    beta = 0.0;

    ArraySetToValue(C,ldc,M,k,0.0);

    c_dusmm_CSC(order, transpose, k, alpha, CSCA, B, ldb, beta, C, ldc);
  
    ArraySave(C,ldc,M,k,"Data/outC4",order);

    /************
      END CASE 4
      ************/

    /***********************************************************
      Case 5 & Case 7:	ColMajor && beta != 0.0 && alpha != 1.0

      *************************************************************/
    
    alpha = 2.25;
    beta = -3.75;

    ArraySetToValue(C,ldc,M,k,1.0);

    c_dusmm_CSC(order, transpose, k, alpha, CSCA, B, ldb, beta, C, ldc);

    ArraySave(C,ldc,M,k,"Data/outC5",order);

    /************
      END CASE 5
      ************/

    /***********************************************************
      Case 8 and Case 10:	RowMajor && beta == 0.0 && alpha == 1.0
      
      *************************************************************/

    order = blas_rowmajor;
    alpha = 1.0;
    beta = 0.0;

    ArraySetToValue(C,ldc,k,M,0.0);

    c_dusmm_CSC(order, transpose, k, alpha, CSCA, B, ldb, beta, C, ldc);

    ArraySave(C,ldc,M,k,"Data/outC8",order);

    /************
      END CASE 8
      ************/

    /***********************************************************
      Case 9 and Case 11:	RowMajor && beta != 0.0 && alpha != 1.0
      
      *************************************************************/

    alpha = 2.25;
    beta = 0.0;
    /*    beta = -3.75;*/

    ArraySetToValue(C,ldc,k,M,1.0);

    c_dusmm_CSC(order, transpose, k, alpha, CSCA, B, ldb, beta, C, ldc);

    ArraySave(C,ldc,M,k,"Data/outC9",order);

    /************
      END CASE 9
      ************/

    /***********************************************************
      Case 12  Column major and B, C submatrices
      
      *************************************************************/
    
    order = blas_colmajor;
    alpha = 1.0;
    ldb = 60;  /* B is a submatrix of BB */
    beta = 0.0;
    ldc = 60;    

    /* initialize CC, which contains the submatrix C */
    ArraySetToValue(CC,ldc,CCM,CCk,0.0);

    c_dusmm_CSC(order, transpose, k, alpha, CSCA, (BB + (ldb * k/2) + N/2), ldb, beta, 
		(CC + (ldc * k) + M), ldc);

    ArraySave((CC + (ldc * k) + M),ldc,M,k,"Data/outC12",order);
    
    /************
      END CASE 12
      ************/

    /***********************************************************
      Case 13  Row major and B, C submatrices
      *************************************************************/

    order = blas_rowmajor;

    ArraySetToValue(CC, ldc, CCM, CCk, 0.0);
    
    c_dusmm_CSC(order, transpose, k, alpha, CSCA, (BB2 + (ldb * N/2) + k/2) , ldb, beta,
		(CC + (ldc * M) + k), ldc);

    ArraySave((CC + (ldc * M) + k),ldc,M,k,"Data/outC13",order);
    
    /************
      END CASE 13
      ************/


    
    /* free memory */
    free(I);
    free(I1);
    free(J);
    free(J1);
    free(val);
    free(B);
    free(C);
    free(BB);
    free(CC);
    free(BB2);

    destroy_DCOO(A);
    destroy_DCOO(A1);
    destroy_DCSC(CSCA);
    destroy_DCSC(CSCA1);

    return 0;

}
