#include "c_blas.h"
#include "spblas_i.h"
#include "csrsm.h"
#include "cscsm.h"

/* lower diagonal */
void c_dussmLD_CSR( enum blas_order_type order, enum blas_trans_type T, const double alpha, 
		    SPBLASI_double_CSR A, int k, double *b, int ldb)
{
  double *val = A->val_;
  int *indx = A->indx_;        /* length = nz */
  int *pntrb = A->pntrb_;      /* length = M */
  int *pntre = A->pntre_;      /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, l, jb, je, index;
  double z;
  double  *pb;
  double valtmp;
  
  SPBLASI_double_CSC CSCA;

  /*
    base one indexing
    */

  if (prop == sparse_blas_one_base) {
    if (order == blas_colmajor) {
      b -= 1;
    }
    else { /* row major base one */
      b -= ldb;
    }
  }

  /* handle transpose case using cscmm */
  if (T == blas_trans)  {

    T = blas_no_trans;
      
    /* create DCSC object */
    CSCA = create_DCSC( m, n, nnz, val, indx, pntrb, pntre, prop);
    
    /* triangular solve */
    c_dussmUD_CSC( order, T, alpha,  CSCA, k, b, ldb);

    free(CSCA);
      
    return;
  }
  

  /* kernels */
  
  if (order == blas_colmajor)
    {
  

      for (l = 0; l != k; l++) {                                
	pb = b;
       
	for (i = 0; i != m; i++) {
	  z = 0;
	  jb = pntrb[i];
	  je = pntre[i];
	  for (j=jb; j<je ; j++) {
	    index = indx[j];
	    if ( index == i ) {
	      valtmp = val[j];
	    } else {
	      z += b[index] * val[j];
	    }
	  }
	  pb[i] = (b[i] - z) / valtmp;
	}
       
	if (alpha != 1.0) {
	  for (i = 0; i != m; i++) {
	    pb[i] *= alpha;
	  }
	}
       
	b += ldb;                               
      }
     
     
    }
  
  
  else if (order == blas_rowmajor)
    {
     
      for (l = 0; l != k; l++) {                                
	pb = b;
	
	for (i = 0; i != m; i++) {
	  z = 0;
	  jb = pntrb[i];
	  je = pntre[i];
	  for (j=jb; j<je ; j++) {
	    index = indx[j];
	    if ( index == i ) {
	      valtmp = val[j];
	    } else {
	      /* z is the sum of t_ik * b_k */
	      z += b[index * ldb] * val[j];
	    }
	  }
	  pb[i * ldb] = (b[i * ldb] - z) / valtmp;
	}
	
	if (alpha != 1.0) {
	  for (i = 0; i != m; i++) {
	    pb[i * ldb] *= alpha;
	  }
	}

	b++;
      } 
    }
}


/* lower unitary diagonal */
void c_dussmLU_CSR( enum blas_order_type order, enum blas_trans_type T, const double alpha, 
		    SPBLASI_double_CSR A, int k, double *b, int ldb)
{
  double *val = A->val_;
  int *indx = A->indx_;        /* length = nz */
  int *pntrb = A->pntrb_;      /* length = M */
  int *pntre = A->pntre_;      /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, l, jb, je, index;
  double z;
  double  *pb;
  double valtmp;
  
  SPBLASI_double_CSC CSCA;

  /*
    base one indexing
    */

  if (prop == sparse_blas_one_base) {
    if (order = blas_colmajor) {
      b = b - 1;
    }
    else { /* row major base one */
      b = b - ldb;
    }
  }

  /* handle transpose case using cscmm */
  if (T == blas_trans)  {

    T = blas_no_trans;
      
    /* create DCSC object */
    CSCA = create_DCSC( m, n, nnz, val, indx, pntrb, pntre, prop);
    
    /* triangular solve */
    c_dussmUU_CSC( order, T, alpha,  CSCA, k, b, ldb);

    free(CSCA);
      
    return;
  }
  

  /* kernels */
  
  if (order == blas_colmajor)
    {
  

      for (l = 0; l != k; l++) {                                
	pb = b;
	
	for (i = 0; i != m; i++) {
	  z = 0;
	  jb = pntrb[i];
	  je = pntre[i];
	  for (j=jb; j<je ; j++) {
	    z += b[indx[j]] * val[j];
	  }
	  pb[i] = b[i] - z;
	}
	
	if (alpha != 1.0) {
	  for (i = 0; i != m; i++) {
	    pb[i] *= alpha;
	  }
	}

	b += ldb;                               
      }

    }
  
  
  else if (order == blas_rowmajor)
    {
      
      for (l = 0; l != k; l++) {                                
	pb = b;
	
	for (i = 0; i != m; i++) {
	  z = 0;
	  jb = pntrb[i];
	  je = pntre[i];
	  for (j=jb; j<je ; j++) {
	    z += b[indx[j] * ldb] * val[j];
	  }
	  pb[i * ldb] = b[i * ldb] - z;
	}
	
	if (alpha != 1.0) {
	  for (i = 0; i != m; i++) {
	    pb[i * ldb] *= alpha ;
	  }
	}

	b++;
      } 
    }
}




/* upper diagonal */
void c_dussmUD_CSR( enum blas_order_type order, enum blas_trans_type T, const double alpha, 
		    SPBLASI_double_CSR A, int k, double *b, int ldb)
{
  double *val = A->val_;
  int *indx = A->indx_;        /* length = nz */
  int *pntrb = A->pntrb_;      /* length = M */
  int *pntre = A->pntre_;      /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, l, jb, je, index;
  double z;
  double  *pb;
  double valtmp;
  
  SPBLASI_double_CSC CSCA;

  /*
    base one indexing
    */

  if (prop == sparse_blas_one_base) {
    if (order = blas_colmajor) {
      b = b - 1;
    }
    else { /* row major base one */
      b = b - ldb;
    }
  }

  /* handle transpose case using cscmm */
  if (T == blas_trans)  {

    T = blas_no_trans;
      
    /* create DCSC object */
    CSCA = create_DCSC( m, n, nnz, val, indx, pntrb, pntre, prop);
    
    /* triangular solve */
    c_dussmLD_CSC( order, T, alpha,  CSCA, k, b, ldb);
     

    free(CSCA);
      
    return;
  }
  

  /* kernels */
  if (order == blas_colmajor)
    {
      for (l = 0; l != k; l++) {                                
	pb = b;
	
	for (i = m-1; i != -1; i--) {
	  
	  z = 0;
	  jb = pntrb[i];
	  je = pntre[i];
	  
	  for (j=jb + 1; j<je ; j++) {
	    z += b[indx[j]] * val[j];
	  }
	  
	  pb[i] = (b[i] - z) / val[jb];
	}
	
	if (alpha != 1.0) {
	  for (i = 0; i != m; i++) {
	    pb[i] *= alpha;
	  }
	}

	b += ldb;                               
      }       
    }
  
  
  else if (order == blas_rowmajor)
    {
      for (l = 0; l != k; l++) {                                
	pb = b;
	
	for (i = m-1; i != -1; i--) {
	  
	  z = 0;
	  jb = pntrb[i];
	  je = pntre[i];
	  
	  for (j=jb + 1; j<je ; j++) {
	    z += b[indx[j] * ldb] * val[j];
	  }
	  
	  pb[i * ldb] = (b[i * ldb] - z) / val[jb];
	}
	
	if (alpha != 1.0) {
	  for (i = 0; i != m; i++) {
	    pb[i * ldb] *= alpha;
	  }
	}

	b++;
      }    
    }

}



/* upper unitary diagonal */
void c_dussmUU_CSR( enum blas_order_type order, enum blas_trans_type T, const double alpha, 
		    SPBLASI_double_CSR A, int k, double *b, int ldb)
{
  double *val = A->val_;
  int *indx = A->indx_;        /* length = nz */
  int *pntrb = A->pntrb_;      /* length = M */
  int *pntre = A->pntre_;      /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, l, jb, je, index;
  double z;
  double  *pb;
  double valtmp;
  
  SPBLASI_double_CSC CSCA;

  /*
    base one indexing
    */

  if (prop == sparse_blas_one_base) {
    if (order = blas_colmajor) {
      b = b - 1;
    }
    else { /* row major base one */
      b = b - ldb;
    }
  }

  /* handle transpose case using cscmm */
  if (T == blas_trans)  {

    T = blas_no_trans;
      
    /* create DCSC object */
    CSCA = create_DCSC( m, n, nnz, val, indx, pntrb, pntre, prop);
    
    /* triangular solve */
    c_dussmLU_CSC( order, T, alpha,  CSCA, k, b, ldb);
     

    free(CSCA);
      
    return;
  }
  

  /* kernels */
  if (order == blas_colmajor)
    {
      for (l = 0; l != k; l++) {                                
	pb = b;
	for (i = m-1; i != -1; i--) {
	  z = 0;
	  jb = pntrb[i];
	  je = pntre[i];
	  for (j=jb; j<je ; j++) {
	    z += b[indx[j]] * val[j];
	  }
	  pb[i] = (b[i] - z);
	}
	if (alpha != 1.0) {
	  for (i = 0; i != m; i++) {
	    pb[i] *= alpha;
	  }
	}

	b += ldb;                               
      }       
    }
  
  
  else if (order == blas_rowmajor)
    {
      for (l = 0; l != k; l++) {                                
	pb = b;
	for (i = m-1; i != -1; i--) {
	  z = 0;
	  jb = pntrb[i];
	  je = pntre[i];
	  for (j=jb; j<je ; j++) {
	    z += b[indx[j] * ldb] * val[j];
	  }
	  pb[i * ldb] = (b[i * ldb] - z);
	}
	if (alpha != 1.0) {
	  for (i = 0; i != m; i++) {
	    pb[i * ldb] *= alpha;
	  }
	}

	b++;
      }    
    }

}

