/*  test_matrix_mult_LD.c  (Mon Aug 30 15:30:51 EDT 1999)
 *  Author:  Galen Wilkerson
 *
 *  See README.txt for usage
 *
 *  Significant parts of this code modified from example_read.c
 *  by Roldan Pozo and Karin Remington
 *
 *  Contact:  galen@cam.nist.gov or pozo@cam.nist.gov
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include "mmio.h"
#include "spblas_i.h"
#include "c_blas.h"

/* initialize a matrix */
void ArraySetToValue(double *A, int ldA, int outer , int inner, double value)
{
  double *p, *pend, *t, *tend;

  pend = A + outer*ldA;
  for (p=A; p<pend; p+=ldA)  {
    tend = p+inner;
    for (t=p; t<tend; *t++ = value);
  }
}

/* print a dense matrix */
void ArrayPrint(double *A, int ldA, int outer , int inner, FILE *output)
{
  double *p, *pend, *t, *tend;
  
  fprintf(output, "%d %d\n", outer, inner);

  pend = A + outer*ldA;

  for (p=A; p<pend; p+=ldA) {
    tend = p+inner;
    for (t=p; t<tend; t++) 
      fprintf(output,"%20.15g ",*t);
    fprintf(output,"\n");
  }
}

/* save a dense matrix */
void ArraySave(double *A, int ldA, int numRows , int numCols, char
	       *outFileName, enum blas_order_type order)
{
  double *p, *pend, *t, *tend;
  FILE *outFile;

  if ((outFile = fopen(outFileName, "w")) == NULL) 
    exit(1);

  fprintf(outFile,"%d %d\n", numRows, numCols);
  
  if (order == blas_colmajor) {
    pend = A + numCols*ldA;
    for (p=A; p<pend; p+=ldA) {
      tend = p + numRows;
      for (t=p; t<tend; t++) 
	fprintf(outFile,"%20.15g ",*t);
      fprintf(outFile,"\n");
    }
  }
  else if (order == blas_rowmajor) {
        
    tend = A + numCols;
    for (t = A; t< tend; t++) {
      pend = t + numRows * ldA;
      for (p = t; p < pend; p += ldA)
	fprintf(outFile,"%20.15g ",*p);
      fprintf(outFile,"\n");
    }
  }
  else 
    exit(1);

  fclose(outFile);  
}

/* read in a dense matrix */
void ArrayRead(double *A, int M , int N, char fileName[] )
{
  FILE *input;
  int i;
  int end = M * N;

  if ((input = fopen(fileName, "r")) == NULL) {
    printf("file not found\n");
    exit(1);
  }

  fscanf(input, "%d %d\n", &M, &N);
  for (i = 0; i < M*N; i++) {
    fscanf(input, "%lg", &A[i]);
  }

  fclose(input);
}

/* find the norm of A - B */
double MatrixNorm(double *A, int lda, double *B, int ldb, int numRows, int numCols, enum blas_order_type order) {
  
  double sum = 0.0;
  double *p, *t, *pb, *tb, *pend, *tend;

  if (order == blas_colmajor) {
    pend = A + numCols*lda;

    pb = B;
    for (p=A; p<pend; p+=lda) {
      tend = p + numRows;
      tb = pb;
      for (t=p; t<tend; t++) {
	sum += pow(*t - *tb, 2);
	tb++;
      }
      pb += ldb;
    }
  }
  else if (order == blas_rowmajor) {
    
    tend = A + numCols;
    tb = B;
    for (t = A; t< tend; t++) {
      pend = t + numRows * lda;
      pb = tb;
      for (p = t; p < pend; p += lda) {
	sum += pow(*p - *pb, 2);
	pb+=ldb;
      }
      tb++;
    }
  }
  else 
    exit(1);

  return (sqrt(sum)/(numRows * numCols));
}


/* copy A to B */
double MatrixCopy(double *A, int lda, double *B, int ldb, int numRows, int numCols, enum blas_order_type order) {
  
  double *p, *t, *pb, *tb, *pend, *tend;

  if (order == blas_colmajor) {
    pend = A + numCols*lda;

    pb = B;
    for (p=A; p<pend; p+=lda) {
      tend = p + numRows;
      tb = pb;
      for (t=p; t<tend; t++) {
	*tb = *t;
	tb++;
      }
      pb += ldb;
    }
  }
  else if (order == blas_rowmajor) {
    
    tend = A + numCols;
    tb = B;
    for (t = A; t< tend; t++) {
      pend = t + numRows * lda;
      pb = tb;
      for (p = t; p < pend; p += lda) {
	*pb = *p;
	pb+=ldb;
      }
      tb++;
    }
  }
  else 
    exit(1);

}


DCSC DCOO_2_DCSC(DCOO A);


int main(int argc, char *argv[])
{
    int ret_code;
    MM_typecode matcode;

    FILE *f, *inB, *inBR, *inBB, *inBB2, *outC0, *outC1, *outC2, *outC3, *outC4, *outC5,
      *outC8, *outC9, *outC12, *outC13;

    FILE *outB12, *outB13;


    /* values to create the compressed sparse row matrix A */
    
    SPBLASI_double_CSC CSCA, CSCA1;

    /* values to create the compressed coordinate matrix A and A1*/
    int M, N, nz;   
    int i, j, temp, *I, *J, *I1, *J1;
    double *val;
    int properties = sparse_blas_zero_base;
    int transpose = blas_no_trans;
    int order = blas_colmajor;
    SPBLASI_double_COO A, A1;

    double alpha = 1.0;
    double *B, *BR, *BB, *BB2;
    double *B0, *B1, *B2, *B3, *B4, *B5;
    int ldb = 30;
    int BBN = 60;
    int BBk = 60;
    int BB2N = 60;
    int BB2k = 60;
    int ldbb = 60;

    /* k == the number of columns of B and C */
    int k = 30;
    
    double beta = 0.0;
    double *C, *CC;
    int ldc = 30;
    int CCM = 60;
    int CCk = 60;
    int ldcc = 60;
    
    char *matrixFileName = "LD_Data/pores_1_LD.mtx";
    char *BfileName = "LD_Data/B30random";
    char *BRfileName = "LD_Data/B30rowmaj";
    char *BBfileName = "LD_Data/BB60random";
    char *BB2fileName = "LD_Data/BB60rowmaj";

    if ((f = fopen(matrixFileName, "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }
    if ((inB = fopen(BfileName, "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }
    if ((inBR = fopen(BRfileName, "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }
    if (mm_read_banner(f, &matcode) != 0)
    {
      printf("Could not process Matrix Market banner.\n");
      exit(1);
    }


    /*  This is how one can screen matrix types if their application */
    /*  only supports a subset of the Matrix Market data types.      */

    if (mm_is_complex(matcode) && mm_is_matrix(matcode) && 
            mm_is_sparse(matcode) )
    {
      printf("Sorry, this application does not support ");
      printf("Market Market type: [%s]\n", mm_typecode_to_str(matcode));
      exit(1);
    }

    /* find out size of sparse matrix .... */

    if ((ret_code = mm_read_mtx_crd_size(f, &M, &N, &nz)) !=0)
        exit(1);


    /* reserve memory for matrix */

    I = (int *) malloc(nz * sizeof(int));
    J = (int *) malloc(nz * sizeof(int));
    val = (double *) malloc(nz * sizeof(double));


    /* reserve memory for 1-base indices */
    I1 = (int *) malloc(nz * sizeof(int));
    J1 = (int *) malloc(nz * sizeof(int));
     
    /* k == the number of columns of B */
    k = N;  /* just make B square */

    /* set sizes of various parameters */
    ldb = N;
    BBN = 2*N;
    BBk = 2*N;
    BB2N = 2*N;
    BB2k = 2*N;
    ldbb = 2*N;
    
    ldc = M;
    CCM = 2 * M;
    CCk = 2 * k;
    ldcc = 2 * ldc;



    /* NOTE: when reading in doubles, ANSI C requires the use of the "l"  */
     /*   specifier as in "%lg", "%lf", "%le", otherwise errors will occur */
     /*  (ANSI C X3.159-1989, Sec. 4.9.6.2, p. 136 lines 13-15)            */

    /* read the compressed coordinate matrix */
     for (i=0; i<nz; i++)
       {
	 fscanf(f, "%d %d %lg\n", &I[i], &J[i], &val[i]);
	 /* adjust from 1-based to 0-based */
	
	 I1[i] = I[i];
	 J1[i] = J[i];

	 I[i]--;  
	 J[i]--;
       }

     if (f !=stdin) fclose(f);


     /************************
      now write out matrix
      mm_write_banner(stdout, matcode);
      mm_write_mtx_crd_size(stdout, M, N, nz);
      for (i=0; i<nz; i++)
      fprintf(stdout, "%d %d %20.19g\n", I[i]+1, J[i]+1, val[i]);
      ************************/    

    /************************************
      create the structure A (zero-based)
      ***********************************/

     A = create_DCOO(M, N, nz, val, I, J, properties);

     /**************************
       convert the DCOO structure 
       A into the DCSC structure CSCA
       
       *************************/  
    
     CSCA = DCOO_2_DCSC(A);

     /************************************
       create the structure A1 (one-based - case 1)
       ***********************************/
     properties = sparse_blas_one_base;
     A1 = create_DCOO(M, N, nz, val, I1, J1, properties);
    
     /**************************
      convert the DCOO structure 
      A into the DCSC structure CSCA
      
      *************************/  

     CSCA1 = DCOO_2_DCSC(A1);
     free(A1);

     /***********************
      allocate memory for B
     **********************/
    
     B = (double *) malloc(N * k * sizeof(double));
     BR = (double *) malloc(N * k * sizeof(double));
     B0 = (double *) malloc(N * k * sizeof(double));
     B1 = (double *) malloc(N * k * sizeof(double));
     B2 = (double *) malloc(N * k * sizeof(double));
     B3 = (double *) malloc(N * k * sizeof(double));
     B4 = (double *) malloc(N * k * sizeof(double));
     B5 = (double *) malloc(N * k * sizeof(double));
    
     /*************************************************
      read the matrix B in col-major order
      ************************************************/
    
     ArrayRead(B, N, k, BfileName);

     /***********************************
      write B to file in row-major order
      **********************************/

     ArraySave(B,ldb,N,k,"LD_Data/B30rowmaj",blas_rowmajor);

     /*************
      allocate C
      **************/
    
     C = (double *) malloc(M * k * sizeof(double));

     /*******************************************************
      create and initialize BB, 
      which contains B as a submatrix (column-major)
      *******************************************************/

     BBN = 2 * N;
     BBk = 2 * k;
    
     BB = (double *) malloc(BBN * BBk * sizeof(double));
     BBfileName = "LD_Data/BB60random";

     if ((inBB = fopen(BBfileName, "r")) == NULL) {
       printf("file not found\n");
       exit(1);
     }
     /* read the matrix BB in col-major order */

     ArrayRead(BB, BBN , BBk, BBfileName);
    
     /* create CC, which contains the submatrix C */
    
     CCM = 2 * M;
     CCk = 2 * k;
    
     CC = (double *) malloc(CCM * CCk * sizeof(double));

     /************************************************
      create and initialize BB2, 
      which contains B as a submatrix (row-major)
      ***********************************************/
    
     BB2N = 2 * N;
     BB2k = 2 * k;
    
     BB2 = (double *) malloc(BB2N * BB2k * sizeof(double));
     if ((inBB2 = fopen(BB2fileName, "r")) == NULL) {
       printf("file not found\n");
       exit(1);
     }

     /*************************************************
      read the matrix BB2 in col-major order
      ************************************************/
    
     ArrayRead(BB2, BB2N , BB2k, BB2fileName);




     /****************************************************************
      Case 0:  vanilla test: alpha == 1.0,  zero base, no transpose, column-major
      
      ****************************************************************/

     ArraySetToValue(C,ldc,M,k,0.0);
     beta = 0;    
    

     printf("\nCSCA:\n");
     DCSC_print(CSCA);


     MatrixCopy(B, ldb, B0, ldb, N,  k,  order);
     /*
     printf("\n B0 :\n");
     ArrayPrint(B0, 30, k, M, stdout);

     printf("\nCase 0\noriginal B:\n");
     ArrayPrint(B, 30, k, M, stdout);
     */
     /* B <- alpha * inv(A) * B     */
     c_dussmLD_CSC(order, transpose, alpha,  CSCA,  k, B, ldb);

     /********************
      save B to file
      *********************/

     ArraySave(B,ldb,M,k,"LD_Data/outB0",order);

     /* check by multiplying:   C <- alpha * A * B + beta * C   */
     c_dusmm_CSC( order, transpose, k, 
		  alpha, CSCA, B, ldb,  beta, C, ldc );
    
     /*
     printf("\nresult C, should equal original B:\n");
     ArrayPrint(C,M,k,M,stdout);
     */

     printf("\nNorm (C - B0): %20.15g \n", MatrixNorm(C,ldc,B0,ldb,M,N, order));
     free(B0);
    
     /*************
      End Case 0
      ************/

    /****************************************************************
      
      Case 1:  base one - column major

    *****************************************************************/
    
     printf("\nCSCA1:\n");
     DCSC_print_i(CSCA1);

     ArrayRead(B, N , k, BfileName);
     MatrixCopy(B, ldb, B1, ldb, N,  k,  order);

     ArraySetToValue(C,ldc,M,k,0.0);

     printf("\nCase 1\noriginal B:\n");
     ArrayPrint(B, 30, k, M, stdout);
     

     c_dussmLD_CSC(order, transpose, alpha,  CSCA1,  k, B, ldb);

     printf("\nresult B:\n");
     ArrayPrint(B, 30, k, M, stdout);
     
     
     ArraySave(C,ldc,M,k,"LD_Data/outB1",order);

     /* check by multiplying:   C <- alpha * A * B + beta * C   */

     c_dusmm_CSC(order, transpose, k, alpha, CSCA1, B, ldb, beta, C, ldc);

     printf("\nresult C, should equal original B:\n");
     ArrayPrint(C,M,k,M,stdout);

     printf("\nNorm (C - B1): %20.15g \n", MatrixNorm(C,ldc,B1,ldb,M,N, order));
     free(B1);

     /************
      END CASE 1
    *************/


   /****************************************************************
      
      Case 2:  base one - row major

    *****************************************************************/


     order = blas_rowmajor;

     ArrayRead(BR, N , k, BRfileName);
     MatrixCopy(BR, ldb, B2, ldb, N,  k,  order);
     ArraySetToValue(C,ldc,M,k,0.0);
 

 
     printf("\nCase 2\noriginal B:\n");
     ArrayPrint(BR, 30, k, M, stdout);
 

     /* B <- alpha * inv(A) * B     */

     c_dussmLD_CSC(order, transpose, alpha,  CSCA1,  k, BR, ldb);

     printf("\nresult B:\n");
     ArrayPrint(BR, 30, k, M, stdout);

     ArraySave(C,ldc,M,k,"LD_Data/outB2",order);

     /* check by multiplying:   C <- alpha * A * B + beta * C   */

     c_dusmm_CSC(order, transpose, k, alpha, CSCA1, BR, ldb, beta, C, ldc);

     printf("\nresult C, should equal original B:\n");
     ArrayPrint(C,M,k,M,stdout);
 
 
     printf("\nNorm (C - B2): %20.15g \n", MatrixNorm(C,ldc,B2,ldb,M,N, order));
     free(B2);

     /************
      END CASE 2
    *************/



   /****************************************************************
     Case 3:  transpose, alpha == 1.0,  zero base,  column-major
      
      ****************************************************************/
    
     order = blas_colmajor;
     transpose = blas_trans;
     beta = 0;    
     ArraySetToValue(C,ldc,M,k,0.0);
     ArrayRead(B, N , k, BfileName);
     MatrixCopy(B, ldb, B3, ldb, N,  k,  order);

     /*
     printf("\ncase 3\noriginal B:\n");
     ArrayPrint(B, 30, k, M, stdout);
     */

     /* B <- alpha * inv(A') * B     */
     c_dussmLD_CSC(order, transpose, alpha,  CSCA,  k, B, ldb);

     /*
     printf("\nresult B:\n");
     ArrayPrint(B, 30, k, M, stdout);
     */

     /********************
      save B to file
      *********************/

     ArraySave(B,ldb,M,k,"LD_Data/outB3",order);

     /* check by multiplying:   C <- alpha * A' * B + beta * C   */
     c_dusmm_CSC( order, transpose, k, 
		  alpha, CSCA, B, ldb,  beta, C, ldc );
    
     /*
     printf("\nresult C, should equal original B:\n");
     ArrayPrint(C,M,k,M,stdout);
     */

     printf("\nNorm (C - B3): %20.15g \n", MatrixNorm(C,ldc,B3,ldb,M,N, order));
     free(B3);
    
     /*************
      End Case 3
      ************/


  /****************************************************************

     Case 4:  submatrix column-major
     no transpose, alpha == 3.25,  zero base,  column-major
      
      ****************************************************************/
     transpose = blas_no_trans;
     alpha = 3.25;
     ldbb = 60;  /* B is a submatrix of BB */
     ldcc = 60;    
    
     beta = 0;
     ArraySetToValue(CC,ldcc,CCM,CCk,0.0);

     MatrixCopy((BB + (ldbb * k/2) + N/2), ldbb, B4, ldb, N,  k,  order);

     /*
     printf("\ncase 4\noriginal B:\n");
     ArrayPrint(B4, ldb, k, M, stdout);
     */

     /* B <- alpha * inv(A) * B     */
     c_dussmLD_CSC(order, transpose, alpha,  CSCA,  k, (BB + (ldbb * k/2) + N/2), ldbb);

     /*
     printf("\nresult B:\n");
     ArrayPrint((BB + (ldbb * k/2) + N/2), ldbb, k, M, stdout);
     */

     /********************
      save B to file
      *********************/

     ArraySave((BB + (ldbb * k/2) + N/2), ldbb, M, k, "LD_Data/outB4", order);

     /* check by multiplying:   C <- (1/alpha) * A * B + beta * C   */
     c_dusmm_CSC( order, transpose, k, 
		  1/alpha, CSCA, (BB + (ldbb * k/2) + N/2) , ldbb,  beta,(CC + (ldcc * k) + M) , ldcc );
    
     /*
     printf("\nresult C, should equal original B:\n");
     ArrayPrint((CC + (ldcc * k) + M),ldcc,k,M,stdout);
     */

     printf("\nNorm (C - B4): %20.15g \n", MatrixNorm((CC + (ldcc * k) + M),ldcc,B4,ldb,M,N, order));
    
     free(B4);
    
     /*************
      End Case 4
      ************/

    /****************************************************************
      Case 5:  submatrix row-major
      no transpose, alpha == -2.75,  zero base,  column-major
      
      ****************************************************************/
     order = blas_rowmajor;    
     alpha = -2.75;
     ldbb = 60;  /* B is a submatrix of BB2 */
     ldcc = 60;    

     MatrixCopy((BB2 + (ldbb * N/2) + k/2), ldbb, B5, ldb, N,  k,  order);

     /*
     printf("\ncase 5\noriginal B5:\n");
     ArrayPrint(B5, ldb, k, M, stdout);
     */
    
     transpose = blas_no_trans;
     beta = 0;
     ArraySetToValue(CC,ldcc,CCM,CCk,0.0);
     
     /*
     printf("\ncase 5\noriginal B:\n");
     ArrayPrint((BB2 + (ldbb * N/2) + k/2), ldbb, k, M, stdout);
     */
 
     /* B <- alpha * inv(A) * B     */
     c_dussmLD_CSC(order, transpose, alpha,  CSCA,  k, (BB2 + (ldbb * N/2) + k/2), ldbb);
     
     /*
     printf("\nresult B:\n");
     ArrayPrint((BB2 + (ldbb * N/2) + k/2), ldbb, k, M, stdout);
     */
     
     /***************
       save B to file
       ***************/
     
     ArraySave((BB2 + (ldbb * N/2) + k/2),ldbb,M,k,"LD_Data/outB5",order);
     
     /* check by multiplying:   C <- (1/alpha) * A * B + beta * C   */
     c_dusmm_CSC( order, transpose, k, 
		  1/alpha, CSCA, (BB2 + (ldbb * N/2) + k/2), ldbb,  beta, (CC + (ldcc * k) + M), ldcc );

     /*     
     printf("\nresult C, should equal original B:\n");
     ArrayPrint((CC + (ldcc * k) + M),ldcc,k,M,stdout);
     */

     printf("\nNorm (C - B5): %20.15g \n", MatrixNorm((CC + (ldcc * k) + M),ldcc,B5,ldb,M,N, order));
     
     free(B5);
     /*************
       End Case 5
       ************/

     /* free memory */
     free(I);
     free(J);
     free(val);
     free(I1);
     free(J1);
     free(B);
     free(BR);
     free(B0);
     free(B1);
     free(B2);
     free(B3);
     free(B4);
     free(B5);
     free(C);
     free(BB);
     free(CC);
     free(BB2);


     destroy_DCOO(A);
     destroy_DCOO(A1);
     destroy_DCSC(CSCA);
     destroy_DCSC(CSCA1);
     return(0);

}
