/*  test_matrix_mult.c  (Mon Aug 30 15:30:51 EDT 1999)
 *  Author:  Galen Wilkerson
 *
 *  See README.txt for usage
 *
 *  Significant parts of this code modified from example_read.c
 *  by Roldan Pozo and Karin Remington
 *
 *  Contact:  galen@cam.nist.gov or pozo@cam.nist.gov
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "mmio.h"
#include "spblas_i.h"
#include "spblas.h"
#include "cblas.h"
#include "csrmm.h"
#include "cscmm.h"


/* initialize a matrix */
void ArraySetToValue(double *A, int ldA, int outer , int inner, double value)
{
  double *p, *pend, *t, *tend;

  pend = A + outer*ldA;
  for (p=A; p<pend; p+=ldA)  {
    tend = p+inner;
    for (t=p; t<tend; *t++ = value);
  }
}

/* print a dense matrix */
void ArrayPrint(double *A, int ldA, int outer , int inner, FILE *output)
{
  double *p, *pend, *t, *tend;
  
  fprintf(output, "%d %d\n", outer, inner);

  pend = A + outer*ldA;

  for (p=A; p<pend; p+=ldA) {
    tend = p+inner;
    for (t=p; t<tend; t++) 
      fprintf(output,"%f ",*t);
    fprintf(output,"\n");
  }
}

/* save a dense vector */
void ArraySave(double *A, int numRows, char *outFileName)
{
  double *p, *pend;
  int numCols = 1;
  FILE *outFile;

  if ((outFile = fopen(outFileName, "w")) == NULL) 
    exit(1);

  fprintf(outFile,"%d %d\n", numRows, numCols);
  
  pend = A + numRows;
  for (p=A; p<pend; p++) 
    fprintf(outFile,"%f ",*p);
  fprintf(outFile,"\n");
  

  fclose(outFile);  
}

/* read in a dense matrix */
void ArrayRead(double *A, int M , int N, char *inFileName)
{
  int i;
  int end = M * N;
  
  FILE *inFile;

  if ((inFile = fopen(inFileName, "r")) == NULL) 
    exit(1);
 
  fscanf(inFile, "%d %d\n", &M, &N);
  for (i = 0; i < M*N; i++) {
    fscanf(inFile, "%lg", &A[i]);
  }

  fclose(inFile);
}


DCSR DCOO_2_DCSR(DCOO A);


int main(int argc, char *argv[])
{
    int ret_code;
    MM_typecode matcode;

    FILE *f, *inX, *inXX, *outX0, *outX1, *outX2, *outX3, *outX4, *outX5, *outX6;
      

    /* values to create the compressed sparse row matrix A */
    
    SPBLASI_double_CSR CSRA, CSRA1;

    /* values to create the compressed coordinate matrix A and A1*/
    int M, N, nz;   
    int incx = 1;
    int i, j, temp, *I, *J, *I1, *J1;
    double *val;
    int properties = blas_zero_base;
    int transpose = CblasNoTrans;
    SPBLASI_double_COO A, A1;

    double alpha = 1.0;
    double *X, *XX;
    double *X0, *X1, *X2,*X3, *X4, *X5, *X6;

    int XXM = 60;
    int XXk = 1;

    /* k == the number of columns of X */
    int k = 1;
    

    /* settings for matrix multiply,to verify trisolve*/
    int order = CblasColMajor;
    double *Y;
    int ldy = M;
    double beta = 0.0;

    char* inXfilename = "LD_Data/X30random";

    if ((f = fopen("LD_Data/pores_1_LD.mtx", "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }

    if (mm_read_banner(f, &matcode) != 0)
    {
      printf("Could not process Matrix Market banner.\n");
      exit(1);
    }


    /*  This is how one can screen matrix types if their application */
    /*  only supports a subset of the Matrix Market data types.      */

    if (mm_is_complex(matcode) && mm_is_matrix(matcode) && 
            mm_is_sparse(matcode) )
    {
      printf("Sorry, this application does not support ");
      printf("Market Market type: [%s]\n", mm_typecode_to_str(matcode));
      exit(1);
    }

    /* find out size of sparse matrix .... */

    if ((ret_code = mm_read_mtx_crd_size(f, &M, &N, &nz)) !=0)
        exit(1);


    /* reserve memory for matrix */

    I = (int *) malloc(nz * sizeof(int));
    J = (int *) malloc(nz * sizeof(int));
    val = (double *) malloc(nz * sizeof(double));


    /* reserve memory for 1-base indices */
    I1 = (int *) malloc(nz * sizeof(int));
    J1 = (int *) malloc(nz * sizeof(int));

    /* NOTE: when reading in doubles, ANSI C requires the use of the "l"  */
    /*   specifier as in "%lg", "%lf", "%le", otherwise errors will occur */
    /*  (ANSI C X3.159-1989, Sec. 4.9.6.2, p. 136 lines 13-15)            */

    /* read the compressed coordinate matrix */
    for (i=0; i<nz; i++)
    {
        fscanf(f, "%d %d %lg\n", &I[i], &J[i], &val[i]);
        /* adjust from 1-based to 0-based */
	
	I1[i] = I[i];
	J1[i] = J[i];

	I[i]--;  
	J[i]--;
    }

    if (f !=stdin) fclose(f);

    /************************/
    /* now write out matrix */
    /************************/

    mm_write_banner(stdout, matcode);
    mm_write_mtx_crd_size(stdout, M, N, nz);
    for (i=0; i<nz; i++)
      fprintf(stdout, "%d %d %20.19g\n", I[i]+1, J[i]+1, val[i]);
    

    /************************************
      create the structure A (zero-based)
      ***********************************/

    A = create_DCOO(M, N, nz, val, I, J, properties);


    /**************************
      convert the DCOO structure 
      A into the DCSR structure CSRA
      
      *************************/  
    
    CSRA = DCOO_2_DCSR(A);
    

    /************************************
      create the structure A1 (one-based - case 1)
      ***********************************/
    
    A1 = create_DCOO(M, N, nz, val, I1, J1, properties);
    
    /**************************
      convert the DCOO structure 
      A into the DCSR structure CSRA
      
      *************************/  
    
    CSRA1 = DCOO_2_DCSR(A);
    
    /***********************
      allocate memory for X
     **********************/
    
    X = (double *) malloc(N * k * sizeof(double));
    

    /*******************************************************
      create and initialize XX, 
      which contains X as a subvector
      *******************************************************/

    XX = (double *) malloc(XXM * XXk * sizeof(double));
    
    /* read the matrix XX */

    ArrayRead(XX, XXM , XXk, "LD_Data/XX60random");
    


    /****************************************************************
      Case 0:  alpha == 1.0, zero base, no transpose
      
      initialize X to 1's
      open the outC0 file
      save C to outC0
      ****************************************************************/
    X0 = (double *) malloc(N * k * sizeof(double));
 
    /*********************************************
      initialize X 
      ********************************************/
    
    ArraySetToValue(X,M,k,M,1.0);
    ArraySetToValue(X0,M,k,M,1.0);
    
    /********************************************

      X <- inv(A) * X
      ********************************************/
    
    printf("\nCSRA:\n");
    DCSR_print(CSRA);
    
    printf("\nCase 0\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    c_dussvLD_CSR( transpose, alpha, CSRA, X, incx );

    printf("\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    /********************
      save X to file
      *********************/

    ArraySave(X,M,"LD_Data/outX0");



    /****************
      multiply X by CSRA, see if it matches X0

    *****************/

    beta = 0.0;
    Y = (double *) malloc(N * k * sizeof(double));
        
    c_dusmm_CSR( order, transpose, k, 
		 alpha, CSRA, X, M,  beta, Y, ldy );
    
    printf("\nY:\n");
    ArrayPrint(Y,M,k,M,stdout);
    
    /*************
      End Case 0
      ************/



    /****************************************************************
      
      Case 1:  transpose (comp row uses comp col triangular solve)

    *****************************************************************/

    transpose = CblasTrans;
    alpha = 1.0;

    /* read X */
    ArrayRead(X, M , k, inXfilename);

    printf("\nCase 1\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    c_dussvLD_CSR( transpose, alpha, CSRA, X, incx );

    printf("\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    ArraySave(X,M,"LD_Data/outX1");

    /************
      END CASE 1
    *************/


    /****************************************************************
      
      Case 2:  alpha == 0.0

      initialize X to non-zero value

      ****************************************************************/
    
    transpose = CblasNoTrans;
    properties = blas_zero_base;
    alpha = 0.0;
    ArraySetToValue(X,M,k,M,1.0);

    printf("\nCase 2\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    c_dussvLD_CSR( transpose, alpha, CSRA, X, incx );
   
    printf("\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    ArraySave(X,M,"LD_Data/outX2");
    
    /***********
      END CASE 2
      ************/

    /****************************************************************
      
      Case 3:  base-one

      initialize X to non-zero value

      ****************************************************************/
    
    properties = blas_one_base;
    alpha = 1.0;

    /* read X */
    ArrayRead(X, M , k, inXfilename);

    printf("\nCase 3\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    c_dussvLD_CSR( transpose, alpha, CSRA1, X, incx );
   
    printf("\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    ArraySave(X,M,"LD_Data/outX3");
    
    /***********
      END CASE 3
      ************/


    /***********************************************************
      Case 4 	alpha == 1.0
 
    *************************************************************/

    properties = blas_zero_base;
    alpha = 1.0;
     
    /* read X from file */
    ArrayRead(X, M , k, inXfilename);
    
    printf("\nCase 4\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    c_dussvLD_CSR( transpose, alpha, CSRA, X, incx );

    printf("\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    ArraySave(X,M,"LD_Data/outX4");

    /************
      END CASE 4
      ************/


    /***********************************************************
      Case 5  alpha != 1.0

     *************************************************************/
    
    alpha = 2.25;

    /* read X from file */
    ArrayRead(X, N , k, inXfilename);

    printf("\nCase 5\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    c_dussvLD_CSR( transpose, alpha, CSRA, X, incx );

    printf("\nX:\n");
    ArrayPrint(X, M, k, M, stdout);

    ArraySave(X,M,"LD_Data/outX5");

    /************
      END CASE 5
      ************/


    /***********************************************************
      Case 6  incx != 1
      
      *************************************************************/

    alpha = 1.0;
    incx = 2;

    c_dussvLD_CSR( transpose, alpha, CSRA, XX, incx );

    /* need to save XX to file, best to save the subarray */
    ArraySave(XX,XXM,"LD_Data/outX6");

    /************
      END CASE 6
      ************/
}

