#include "c_blas.h"
#include "spblas_i.h"
#include "microblas.h"
#include "dbsrmmi.h"
#include "dbscmmi.h"

void c_dusmv_BSR(enum blas_trans_type T, const double alpha, 
		DBSR A, const double *x,
		int incx, const double beta, double *y, int incy)
{
	c_dusmm_BSR(blas_colmajor, T, 1, alpha, A, x, incx, beta, y, incy);
}


/* NOTE: we assume that ordering for the blocks is the same
as the ordering for the dense arguments (B, C) */

void c_dusmm_BSR(enum blas_order_type order, 
		enum blas_trans_type T, int k,
		double alpha, DBSR A, const double *B,
		int ldB, double beta, double *C, int ldC)
{

	if (T == blas_trans)
		dbscmm(A->nb_, order, blas_trans, order, 
			A->Nnb_, A->Mnb_, k, alpha, A->pntrb_, A->pntre_, A->val_, 
			A->indx_, B, ldB, beta, C, ldC);
	else
		dbsrmm(A->nb_, order, blas_no_trans, order, 
			A->Mnb_, A->Nnb_, k, alpha, A->pntrb_, A->pntre_, A->val_, 
			A->indx_, B, ldB, beta, C, ldC);
}



void dbsrmm(int nb, enum sparse_blas_base_type base, 
		enum blas_trans_type transblock, 
		enum blas_order_type order, int Mnb, int Nnb, int K, double alpha, 
		const int *pntrb, const int *pntre, const double *val,
		const int *indx, const double *B, int ldB, double beta,
		double *C, int ldC)
{
		if (order == blas_colmajor)
		{
			if (base == sparse_blas_one_base)
			{
				B -= nb;
				C -= nb;
			}
			dbsrmm_col(nb, transblock, Mnb, K, pntrb, pntre, val, indx,
				alpha, B, ldB, beta, C, ldC);
		}
		else /* row-major */
		{
			if (base == sparse_blas_one_base)
			{
				B -= ldB * nb; 
				C -= ldC * nb;
			}
			dbsrmm_row(nb, transblock, Mnb, K, pntrb, pntre, val, indx,
				alpha, B, ldB, beta, C, ldC);
		}
	
}


/* C = alpha * A*B + beta*C;  A is compressed-row,                     */
/*							A's blocks, B, C are row-major             */   


void dbsrmm_row(int nb, enum blas_trans_type transblock, int Mnb, 
			int K, const int *pntrb, const int *pntre, 
			const double *val, const int *indx, double alpha, 
			const double *B, int ldB, double beta, double *C, int ldC)

{
	int i;
	int nb2 = nb*nb;

	/* C is a single column of (nb x K) blocks */
	for (i=0; i<Mnb; i++)	/* each block row of C */
	{
		double *Crowi = &C[ldC * i * nb];

		/* compute the ith nb x nb block */
		ublas_bspdotab_row(nb, transblock, 
			Mnb, K, alpha, pntrb[i], pntre[i], val, indx,
			B, ldB, beta, Crowi, ldC);
	}

}



/* C = alpha * A*B + beta*y;  A is compressed-row, B&C are col-major  */   


void dbsrmm_col(int nb, enum blas_trans_type transblock, int Mnb, int K, 
			const int *pntrb, const int *pntre, 
			const double *val, const int *indx, double alpha, 
			const double *B, int ldB, double beta, double *C, int ldC)

{
	int i,k;

	for (i=0; i<Mnb; i++)	/* each column of C */
	{
		const double  *Bi = &B[i*nb];
		double  *Ci = &C[i*nb];
		for (k=0; k<Mnb; k++)
		{
			ublas_bspdotab_col(nb, transblock, 
			Mnb, k, alpha,  pntrb[i], pntre[i], val, indx, Bi, ldB,
						beta, &Ci[k], ldC);
		}
	}
}


void dbscmm(int nb, enum sparse_blas_base_type base, 
	    enum blas_trans_type transblock, 
	    enum blas_order_type order, int Mnb, int Nnb, int K, double alpha, 
	    const int *pntrb, const int *pntre, const double *val,
	    const int *indx, const double *B, int ldB, double beta,
	    double *C, int ldC) {}


