/*  test_matrix_mult.c  (Wed Aug  4 15:47:25 EDT 1999)
 *  Author:  Galen Wilkerson
 *
 *  See README.txt for usage
 *
 *  Significant parts of this code modified from example_read.c
 *  by Roldan Pozo and Karin Remington
 *
 *  Contact:  galen@cam.nist.gov or pozo@cam.nist.gov
 */

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "mmio.h"
#include "spblas_i.h"
#include "c_blas.h"

/* initialize a matrix */
void ArraySetToValue(double *A, int ldA, int outer , int inner, double value)
{
  double *p, *pend, *t, *tend;

  pend = A + outer*ldA;
  for (p=A; p<pend; p+=ldA)  {
    tend = p+inner;
    for (t=p; t<tend; *t++ = value);
  }
}

/* print a dense matrix */
void ArrayPrint(double *A, int ldA, int outer , int inner, FILE *output)
{
  double *p, *pend, *t, *tend;
  
  fprintf(output, "%d %d\n", outer, inner);

  pend = A + outer*ldA;

  for (p=A; p<pend; p+=ldA) {
    tend = p+inner;
    for (t=p; t<tend; t++) 
      fprintf(output,"%f ",*t);
    fprintf(output,"\n");
  }
}

/* save a dense matrix */
void ArraySave(double *A, int ldA, int numRows , int numCols, char
	       *outFileName, enum blas_order_type order)
{
  double *p, *pend, *t, *tend;
  FILE *outFile;

  if ((outFile = fopen(outFileName, "w")) == NULL) 
    exit(1);

  fprintf(outFile,"%d %d\n", numRows, numCols);
  
  if (order == blas_colmajor) {
    pend = A + numCols*ldA;
    for (p=A; p<pend; p+=ldA) {
      tend = p + numRows;
      for (t=p; t<tend; t++) 
	fprintf(outFile,"%f ",*t);
      fprintf(outFile,"\n");
    }
  }
  else if (order == blas_rowmajor) {
        
    tend = A + numCols;
    for (t = A; t< tend; t++) {
      pend = t + numRows * ldA;
      for (p = t; p < pend; p += ldA)
	fprintf(outFile,"%f ",*p);
      fprintf(outFile,"\n");
    }
  }
  else 
    exit(1);

  fclose(outFile);  
}

/* read in a dense matrix */
void ArrayRead(double *A, int M , int N, FILE *input)
{
  int i;
  int end = M * N;
  
  fscanf(input, "%d %d\n", &M, &N);
  for (i = 0; i < M*N; i++) {
    fscanf(input, "%lg", &A[i]);
  }
}


int Read_mtx_size(FILE *f, int *nb, int *M, int *N, int *nz )
{
    char line[MM_MAX_LINE_LENGTH];
    int num_items_read;

    /* set return null parameter values, in case we exit with errors */
    *M = *N = *nz = 0;

    /* now continue scanning until you reach the end-of-comments */
    do 
    {
        if (fgets(line,MM_MAX_LINE_LENGTH,f) == NULL) 
            return MM_PREMATURE_EOF;
    }while (line[0] == '%');

    /* line[] is either blank or has nb, M,N, nz */
    if (sscanf(line, "%d %d %d", nb, M, N, nz) == 3)
        return 0;
        
    else
    do
    { 
        num_items_read = fscanf(f, "%d %d %d", M, N, nz); 
        if (num_items_read == EOF) return MM_PREMATURE_EOF;
    }
    while (num_items_read != 3);

    return 0;
}


/* convert the DBCO to DCOO and save */
void Save_dbco(char *fname, int M, int N, int nzb, int nb, double *val, int *indi, int *indj, int order) {

  
  int *I = (int *) malloc(nzb * nb * nb * sizeof(int));
  int *J = (int *) malloc(nzb * nb * nb * sizeof(int));
  MM_typecode matcode;
  mm_initialize_typecode(&matcode);
  mm_set_matrix(&matcode);
  mm_set_coordinate(&matcode);
  mm_set_real(&matcode);

  /* fill I and J with the correct index values */
  dbco2coo(order,  nzb,  nb, val, indi, indj, I, J);

  mm_write_mtx_crd(fname,  M,  N,  nzb*nb*nb,  I, J, val, matcode);
}

DCSR DCOO_2_DCSR(DCOO A);


int main(int argc, char *argv[])
{
    int ret_code;
    MM_typecode matcode;

    FILE *f, *inB, *inBB, *inBB2, *outC0, *outC1, *outC2, *outC3, *outC4, *outC5,
      *outC8, *outC9, *outC12, *outC13;

    FILE *outB12, *outB13;

    char *AfileName, *A1fileName;

    /* values to create the compressed coordinate matrix A */
    int M, N;
    int Mnb, Nnb;
    int nb = 5;
    int  nz;
    int nzb;
    int Mb, Nb;
    int i, j, temp, *I, *J, *I1, *J1;
    double *val, *temp_val;
    int properties;
    int transpose = blas_no_trans;
    int order = blas_colmajor;
    int block_ordering;


    SPBLASI_double_COO A;
    SPBLASI_double_CSR CSRA;

    int *indx, *pntrb, *pntre;

    double alpha = 1.0;
    double *B, *BB;
    int ldb = 150;

    int BBN = 300;
    int BBk = 300;
    int ldbb = 2 * ldb;

    /* k == the number of columns of B and C */
    int k = 150;
    
    double beta = 0.0;
    double *C, *CC;
    int ldc = 150;
    int CCM = 300;
    int CCk = 300;
    int ldcc = 2 * ldc;
   
    double rand_val;
    double randLimit = 1e+04;

    time_t *tp;
    
    /* seed the rand() function 

    srand((double) time(tp));
    */
    printf("start\n");


    if ((f = fopen("Data/pores_1.mtx", "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }
    if ((inB = fopen("Data/B150random", "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }
    if ((inBB = fopen("Data/BB300random", "r")) == NULL) {
      printf("file not found\n");

      exit(1);
    }
    if (mm_read_banner(f, &matcode) != 0)
    {
      printf("Could not process Matrix Market banner.\n");
      exit(1);
    }




    /*  This is how one can screen matrix types if their application */
    /*  only supports a subset of the Matrix Market data types.      */

    if (mm_is_complex(matcode) && mm_is_matrix(matcode) && 
            mm_is_sparse(matcode) )
    {
      printf("Sorry, this application does not support ");
      printf("Market Market type: [%s]\n", mm_typecode_to_str(matcode));
      exit(1);
    }

    /* find out size of sparse matrix .... */

    if ((ret_code = mm_read_mtx_crd_size(f, &M, &N, &nz)) !=0)
        exit(1);


    printf("read size\n");

    /* reserve memory for matrix */
    I = (int *) malloc(nz * sizeof(int));
    J = (int *) malloc(nz * sizeof(int));
    val = (double *) malloc(nz * nb * nb * sizeof(double));
    temp_val = (double *) malloc(nz * sizeof(double));

    /* reserve memory for 1-base indices */
    I1 = (int *) malloc(nz * sizeof(int));
    J1 = (int *) malloc(nz * sizeof(int));

    /* NOTE: when reading in doubles, ANSI C requires the use of the "l"  */
    /*   specifier as in "%lg", "%lf", "%le", otherwise errors will occur */
    /*  (ANSI C X3.159-1989, Sec. 4.9.6.2, p. 136 lines 13-15)            */

    /* read the compressed coordinate matrix */
    for (i=0; i<nz; i++)
    {
        fscanf(f, "%d %d %lg\n", &I[i], &J[i], &temp_val[i]);
        /* adjust from 1-based to 0-based */
	
	I1[i] = I[i];
	J1[i] = J[i];

	I[i]--;  
	J[i]--;
    }

    if (f !=stdin) fclose(f);


    /************************************
      create the structure A (zero-based)
      ***********************************/

    A = create_DCOO(M, N, nz, temp_val, I, J, properties);
    printf("DCOO A\n");
    DCOO_print(A);


    /**************************
      convert the DCOO structure 
      A into the DCSR structure CSRA
      
      *************************/  
    
    CSRA = DCOO_2_DCSR(A);
    printf("\nCSRA:\n");
    DCSR_print(CSRA);
    

    /************************************
      fill in the the other values for BCO
      ***********************************/
    block_ordering = 1;
    properties = sparse_blas_zero_base;
    Mnb = M;
    Nnb = N;
    M = M * nb;
    N = N * nb;
    nzb = nz;
    nz = nz * nb * nb;


    /* fill val with random values */
    for (i = 0; i < nz; i++) {
      rand_val = (rand() / (RAND_MAX / randLimit + 1));
      if ((rand()/(double) RAND_MAX) < .5 )
	rand_val *= -1;

      val[i] = rand_val;
    }


    /***********************************
      save as COO format
      ***********************************/

    A1fileName = "Data/outA1";

    /* iterate through I1 and J1, 
       multiplying by nb to allow for blocksize*/
    for (i=0; i<nzb ;i++) {
      I1[i] = I1[i] * nb - (nb -1);
      J1[i] = J1[i] * nb - (nb -1);
    }

    Save_dbco(A1fileName, M,  N,  nzb, nb, val, I1, J1, order);

    /***********************************
      create BSR matrix

    **********************************/

    
    indx = CSRA->indx_;
    pntrb = CSRA->pntrb_;
    pntre = CSRA->pntre_;
    


    /***********************
      allocate memory for B
     **********************/
    
    B = (double *) malloc(N * k * sizeof(double));
    
    /*************************************************
      read (and print) the matrix B in col-major order
      ************************************************/
    
    ArrayRead(B, N, k, inB);

    /*************
      allocate C
      **************/
    
    C = (double *) malloc(M * k * sizeof(double));

    /*******************************************************
      create and initialize BB, 
      which contains B as a submatrix (column-major - case 12)
      *******************************************************/

    BBN = 2 * N;
    BBk = 2 * k;
    
    BB = (double *) malloc(BBN * BBk * sizeof(double));
    
    /* read the matrix BB in col-major order */

    ArrayRead(BB, BBN , BBk, inBB);
    
    /* create CC, which contains the submatrix C */
    
    CCM = 2 * M;
    CCk = 2 * k;
    
    CC = (double *) malloc(CCM * CCk * sizeof(double));


    /****************************************************************

      Case 0
      transpose = no trans
      alpha = 3.25
      nb = 5
      nb_ordering = col
      beta = -2.75
      ldb = M;
      ldc = N;
      prop = base 0
      order = col

      ****************************************************************/


    /*********************************************
      initialize C in column-major order
      ********************************************/
    
    ArraySetToValue(C, ldc, M, k, 0.0);
    
    /********************************************
      multiply the matrix A by B

      C <- A * B
      ********************************************/

    /* block multiply */
    dbsrmm( nb,  properties, transpose, order,  Mnb,  Nnb,  k,  alpha, 
	    pntrb, pntre, val, indx, B, ldb, beta, C, ldc);

    /********************
      save C to file
      *********************/

    ArraySave(C,ldc,M,k,"Data/outC0",order);

    /*************
      End Case 0
      ************/

 
    /****************************************************************

      Case 1
      transpose = trans
      alpha = 3.25
      nb = 5
      nb_ordering = row
      beta = -2.75
      ldb = 2*M
      ldc = 2*N
      prop = base 1
      order = row

      ****************************************************************/
    transpose = blas_trans;
    order = blas_rowmajor;
    
    ArraySetToValue(CC, ldcc, CCM, CCk, 0.0);
    
    /* block multiply */
    dbsrmm( nb,  properties, transpose, order,  Mnb,  Nnb,  k,  alpha, 
	    pntrb, pntre, val, indx, BB, ldbb, beta, CC, ldcc);
    
    ArraySave(CC,ldcc,CCM,CCk,"Data/outC1",order);
    
    /************
      END CASE 1
    *************/


    /* free memory */
    free(I);
    free(I1);
    free(J);
    free(J1);
    free(val);
    free(temp_val);
    free(B);
    free(C);
    free(BB);
    free(CC);

    destroy_DCOO(A);
    destroy_DCSR(CSRA);

}
