#include "cblas.h"
#include "spblas_i.h"
#include "spblas.h"
#include "csrsm.h"
#include "cscsm.h"

/* lower diagonal */

void c_dussmLD_CSC( enum CBLAS_ORDER order, enum CBLAS_TRANSPOSE T, const double alpha, 
		    SPBLASI_double_CSC A, int k, double *b, int ldb)

{
  double *val = A->val_;
  int *indx = A->indx_;        /* length = nz */
  int *pntrb = A->pntrb_;      /* length = M */
  int *pntre = A->pntre_;      /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, l, jb, je, index;
  double z;
  double  *pb;
  double valtmp;
  
  SPBLASI_double_CSR CSRA;


  /*
    base one indexing
   */

  if (prop == blas_one_base) {
    if (order = CblasColMajor) {
      b = b - 1;
    }
    else { /* row major base one */
      b = b - ldb;
    }
  }

  /* handle transpose case using cscmm */
  if (T == CblasTrans)  {

      T = CblasNoTrans;
      
      /* create DCSR object */
      CSRA = create_DCSR( m, n, nnz, val, indx, pntrb, pntre, prop);
      
      /* triangular solve */
      c_dussmUD_CSR( order, T, alpha,  CSRA, k, b, ldb);

      free(CSRA);
      
      return;
  }
  

  /* kernel */
  if (order == CblasColMajor)
    {

      for (l = 0; l != k; l++) {                                
	pb = b;
	for (i = 0; i != m; i++) {
	  jb = pntrb[i];
	  je = pntre[i];
	  z =  pb[i] / val[jb];
	  pb[i] = z;
	  for (j=jb+1;j<je;j++) {
	    b[indx[j]] -= z*val[j];
	  }
	}
	if (alpha != 1.0) 
	  for (i=0;i!=m;i++) {                  
	    pb[i] *= alpha ;            
	  }                                     

	b += ldb;                               
      }                 
    }

  else if (order == CblasRowMajor)
    {
      
      for (l = 0; l != k; l++) {                                
	pb = b;
	for (i = 0; i != m; i++) {
	  jb = pntrb[i];
	  je = pntre[i];
	  z =  pb[i * ldb] / val[jb];
	  pb[i * ldb] = z;
	  for (j=jb+1;j<je;j++) {
	    b[indx[j] * ldb] -= z*val[j];
	  }
	}
	if (alpha != 1.0) 
	  for (i=0;i!=m;i++) {                  
	    pb[i * ldb] *= alpha ;            
	  }
                                     
	b++;
      }                 
    }
}






/* lower unitary  */
void c_dussmLU_CSC( enum CBLAS_ORDER order, enum CBLAS_TRANSPOSE T, const double alpha, 
		    SPBLASI_double_CSC A, int k, double *b, int ldb)

{
  double *val = A->val_;
  int *indx = A->indx_;        /* length = nz */
  int *pntrb = A->pntrb_;      /* length = M */
  int *pntre = A->pntre_;      /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, l, jb, je, index;
  double z;
  double  *pb;
  double valtmp;
  
  SPBLASI_double_CSR CSRA;


  /*
    base one indexing
   */

  if (prop == blas_one_base) {
    if (order = CblasColMajor) {
      b = b - 1;
    }
    else { /* row major base one */
      b = b - ldb;
    }
  }

  /* handle transpose case using cscmm */
  if (T == CblasTrans)  {

      T = CblasNoTrans;
      
      /* create DCSR object */
      CSRA = create_DCSR( m, n, nnz, val, indx, pntrb, pntre, prop);
      
      /* triangular solve */
      c_dussmUU_CSR( order, T, alpha,  CSRA, k, b, ldb);

      free(CSRA);
      
      return;
  }
  

  /* kernel */
  if (order == CblasColMajor)
    {

      for (l = 0; l != k; l++) {                                
	pb = b;
	for (i = 0; i != m; i++) {
	  jb = pntrb[i];
	  je = pntre[i];
	  z =  pb[i];
	  for (j=jb+1;j<je;j++) {
	    b[indx[j]] -= z*val[j];
	  }
	}
	if (alpha != 1.0) 
	  for (i=0;i!=m;i++) {                  
	    pb[i] *= alpha ;            
	  }                                     

	b += ldb;                               
      }                 
    }

  else if (order == CblasRowMajor)
    {
      
      for (l = 0; l != k; l++) {                                
	pb = b;
	for (i = 0; i != m; i++) {
	  jb = pntrb[i];
	  je = pntre[i];
	  z =  pb[i * ldb];
	  for (j=jb+1;j<je;j++) {
	    b[indx[j] * ldb] -= z*val[j];
	  }
	}
	if (alpha != 1.0) 
	  for (i=0;i!=m;i++) {                  
	    pb[i * ldb] *= alpha ;            
	  }
                                     
	b++;
      }                 
    }
}






/* upper diagonal */
void c_dussmUD_CSC( enum CBLAS_ORDER order, enum CBLAS_TRANSPOSE T, const double alpha, 
		    SPBLASI_double_CSC A, int k, double *b, int ldb)
{
  double *val = A->val_;
  int *indx = A->indx_;        /* length = nz */
  int *pntrb = A->pntrb_;      /* length = M */
  int *pntre = A->pntre_;      /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, l, jb, je, index;
  double z;
  double  *pb;
  double valtmp;
  
  SPBLASI_double_CSR CSRA;

  /*
    base one indexing
   */

  if (prop == blas_one_base) {
    if (order = CblasColMajor) {
      b = b - 1;
    }
    else { /* row major base one */
      b = b - ldb;
    }
  }

  /* handle transpose case using cscmm */
  if (T == CblasTrans)  {

      T = CblasNoTrans;
      
      /* create DCSR object */
      CSRA = create_DCSR( m, n, nnz, val, indx, pntrb, pntre, prop);
      
      /* triangular solve */

      c_dussmLD_CSR( order, T, alpha,  CSRA, k, b, ldb);

      free(CSRA);
      
      return;
  }
  

  /* kernel */
  if (order == CblasColMajor)
    {

      for (l = 0; l != k; l++) {                                
	pb = b;
	for (i = m-1; i != -1; i--) {
	  jb = pntrb[i];
	  je = pntre[i]-1;
	  z =  pb[i] / val[je];
	  pb[i] = z;
	  for (j = jb; j < je; j++) {
	    b[indx[j]] -= z * val[j];
	  }
	}

	if (alpha != 1.0) 
	  for (i=0;i!=m;i++) {                  
	    pb[i] *= alpha ;            
	  }                                     

	b += ldb;                               
      }                 
    }

  else if (order == CblasRowMajor)
    {
      
      for (l = 0; l != k; l++) {                                
	pb = b;
	for (i = m-1; i != -1; i--) {
	  jb = pntrb[i];
	  je = pntre[i]-1;
	  z =  pb[i * ldb] / val[je];
	  pb[i * ldb] = z;
	  for (j = jb; j < je; j++) {
	    b[indx[j] * ldb] -= z * val[j];
	  }
	}

	if (alpha != 1.0) 
	  for (i=0;i!=m;i++) {                  
	    pb[i * ldb] *= alpha ;            
	  }
                                     
	b++;
      }                 
    }
}




/* upper diagonal */
void c_dussmUU_CSC( enum CBLAS_ORDER order, enum CBLAS_TRANSPOSE T, const double alpha, 
		    SPBLASI_double_CSC A, int k, double *b, int ldb)
{
  double *val = A->val_;
  int *indx = A->indx_;        /* length = nz */
  int *pntrb = A->pntrb_;      /* length = M */
  int *pntre = A->pntre_;      /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, l, jb, je, index;
  double z;
  double  *pb;
  double valtmp;
  
  SPBLASI_double_CSR CSRA;

  /*
    base one indexing
   */

  if (prop == blas_one_base) {
    if (order = CblasColMajor) {
      b = b - 1;
    }
    else { /* row major base one */
      b = b - ldb;
    }
  }

  /* handle transpose case using cscmm */
  if (T == CblasTrans)  {

      T = CblasNoTrans;
      
      /* create DCSR object */
      CSRA = create_DCSR( m, n, nnz, val, indx, pntrb, pntre, prop);
      
      /* triangular solve */
      c_dussmLU_CSR( order, T, alpha,  CSRA, k, b, ldb);

      free(CSRA);
      
      return;
  }
  

  /* kernel */
  if (order == CblasColMajor)
    {

      for (l = 0; l != k; l++) {                                
	pb = b;
	for (i = m-1; i != -1; i--) {
	  jb = pntrb[i];
	  je = pntre[i];
	  z =  pb[i];
	  for (j = jb; j < je; j++) {
	    b[indx[j]] -= z * val[j];
	  }
	}

	if (alpha != 1.0) 
	  for (i=0;i!=m;i++) {                  
	    pb[i] *= alpha ;            
	  }                                     

	b += ldb;                               
      }                 
    }

  else if (order == CblasRowMajor)
    {
      
      for (l = 0; l != k; l++) {                                
	pb = b;
	for (i = m-1; i != -1; i--) {
	  jb = pntrb[i];
	  je = pntre[i];
	  z =  pb[i * ldb];
	  for (j = jb; j < je; j++) {
	    b[indx[j] * ldb] -= z * val[j];
	  }
	}

	if (alpha != 1.0) 
	  for (i=0;i!=m;i++) {                  
	    pb[i * ldb] *= alpha ;            
	  }
                                     
	b++;
      }                 
    }
}



