#include <stdio.h>
#include <string.h>
#include "c_blas.h"
#include "spblas_i.h"
#include "cscsv.h"


void c_dussvLD_CSR( 
		   enum blas_trans_type T, 
		   const double alpha, 
		   SPBLASI_double_CSR A, 
		   double *x, 
		   int incx )
     
{
  double *val = A->val_;
  int *indx = A->indx_;             /* length = nz */
  int *pntrb = A->pntrb_;            /* length = M */
  int *pntre = A->pntre_;            /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, jb, je, index;
  double *px=x;
  double z; 
  double valtmp;

  SPBLASI_double_CSC CSCA;


  /* handle transpose case using cscsv */
  if (T == blas_trans)  {

      T = blas_no_trans;
      
      /* create DCSC object */
      CSCA = create_DCSC( m, n, nnz, val, indx, pntrb, pntre, prop);
      
      /* triangular solve */
      c_dussvUD_CSC( T, alpha, CSCA, x, incx);
  
      free(CSCA);
      return;
  }
  


  /* early return for silly argument values */
  if (alpha == 0)
    {
      for (i = 0; i < m; i++) 
	x[i] = 0.0;
      return;
    }

  if (prop == sparse_blas_one_base) {
    x = x - incx;
    px = x;
  }

  for (i=0;i!=m;i++) {
    z = 0;
    jb = pntrb[i];
    je = pntre[i];
    for (j=jb;j<je;j++) {
      index = indx[j];
      if ( index == i ) {
	valtmp = val[j];
      } else {
	z += x[index * incx] * val[j];
      }
    }
    px[i * incx] = (x[i * incx] - z) / valtmp;
  }
  
  
  if (alpha != 1.0)
    for (i = 0; i < (m * incx); i+=incx) 
      px[i] *= alpha;
    
}





void c_dussvLU_CSR( 
		   enum blas_trans_type T, 
		   const double alpha, 
		   SPBLASI_double_CSR A, 
		   double *x, 
		   int incx )
     
{
  double *val = A->val_;
  int *indx = A->indx_;             /* length = nz */
  int *pntrb = A->pntrb_;            /* length = M */
  int *pntre = A->pntre_;            /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, jb, je, index;
  double *px=x;
  double z; 
  double valtmp;

  SPBLASI_double_CSC CSCA;


  /* handle transpose case using cscsv */
  if (T == blas_trans)  {

      T = blas_no_trans;
      
      /* create DCSC object */
      CSCA = create_DCSC( m, n, nnz, val, indx, pntrb, pntre, prop);
      
      /* triangular solve */
      c_dussvUU_CSC( T, alpha, CSCA, x, incx);
  
      free(CSCA);
      
      return;
  }

  /* early return for silly argument values */
  if (alpha == 0)
    {
      for (i = 0; i < m; i++) 
	x[i] = 0.0;
      return;
    }

  if (prop == sparse_blas_one_base) {
    x = x - incx;
    px = x;
  }

  for (i=0;i!=m;i++) {
    z = 0;
    jb = pntrb[i];
    je = pntre[i];
    for (j=jb;j<je;j++) {
	z += x[indx[j] * incx] * val[j];
    }
    px[i * incx] = x[i * incx] - z;
  }
  
  if (alpha != 1.0)
    for (i = 0; i < (m * incx); i+=incx) 
      px[i] *= alpha;
    
}



void c_dussvUD_CSR(enum blas_trans_type T, 
		   const double alpha, 
		   SPBLASI_double_CSR A, 
		   double *x, 
		   int incx )
     
{
  double *val = A->val_;
  int *indx = A->indx_;             /* length = nz */
  int *pntrb = A->pntrb_;            /* length = M */
  int *pntre = A->pntre_;            /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, jb, je, index;
  double *px=x;
  double z; 
  double valtmp;

  SPBLASI_double_CSC CSCA;

  /* handle transpose case using cscsv */
  if (T == blas_trans)  {

      T = blas_no_trans;
      
      /* create DCSC object */
      CSCA = create_DCSC( m, n, nnz, val, indx, pntrb, pntre, prop);
      
      /* triangular solve */
      c_dussvLD_CSC( T, alpha, CSCA, x, incx);
      free(CSCA);
      
      return;
  }

  /* early return for silly argument values */
  if (alpha == 0)
    {
      for (i = 0; i < m; i++) 
	x[i] = 0.0;
      return;
    }

  if (prop == sparse_blas_one_base) {
    x = x - incx;
    px = x;
  }

  /* this kernel corresponds to: x <- alpha * inv(A) * x    */
  for (i=m-1;i!=-1; i--) {
    z = 0;
    jb = pntrb[i];
    je =  pntre[i];
    for (j=jb+1; j<je; j++) {
      z += x[indx[j] * incx] * val[j];
    }
    px[i * incx] = (x[i * incx] - z) / val[jb];
  }

  if (alpha != 1.0)
    for (i = 0; i < (m * incx); i+=incx) 
      px[i] *= alpha;
}




void c_dussvUU_CSR(enum blas_trans_type T, 
		   const double alpha, 
		   SPBLASI_double_CSR A, 
		   double *x, 
		   int incx )
     
{
  double *val = A->val_;
  int *indx = A->indx_;             /* length = nz */
  int *pntrb = A->pntrb_;            /* length = M */
  int *pntre = A->pntre_;            /* length = M */
  int nnz = A->nz_;
  int prop = A->prop_;
  int m = A->M_;
  int n = A->N_;

  int i, j, jb, je, index;
  double *px=x;
  double z; 
  double valtmp;

  SPBLASI_double_CSC CSCA;

  /* handle transpose case using cscsv */
  if (T == blas_trans)  {

      T = blas_no_trans;
      
      /* create DCSC object */
      CSCA = create_DCSC( m, n, nnz, val, indx, pntrb, pntre, prop);
      
      /* triangular solve */
      c_dussvLU_CSC( T, alpha, CSCA, x, incx);
      free(CSCA);
      
      return;
  }

  /* early return for silly argument values */
  if (alpha == 0)
    {
      for (i = 0; i < m; i++) 
	x[i] = 0.0;
      return;
    }

  if (prop == sparse_blas_one_base) {
    x = x - incx;
    px = x;
  }

  /* this kernel corresponds to: x <- alpha * inv(A) * x    */
  for (i=m-1;i!=-1; i--) {
    z = 0;
    jb = pntrb[i];
    je =  pntre[i];
    for (j=jb; j<je; j++) {
      z += x[indx[j] * incx] * val[j];
    }
    px[i * incx] = x[i * incx] - z;
  }

  if (alpha != 1.0)
    for (i = 0; i < (m * incx); i+=incx) 
      px[i] *= alpha;
}
