/*  test_matrix_mult.c  (Wed Aug  4 15:47:25 EDT 1999)
 *  Author:  Galen Wilkerson
 *
 *  See README.txt for usage
 *
 *  Significant parts of this code modified from example_read.c
 *  by Roldan Pozo and Karin Remington
 *
 *  Contact:  galen@cam.nist.gov or pozo@cam.nist.gov
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "mmio.h"
#include "spblas_i.h"
#include "spblas.h"
#include "cblas.h"

/* initialize a matrix */
void ArraySetToValue(double *A, int ldA, int outer , int inner, double value)
{
  double *p, *pend, *t, *tend;

  pend = A + outer*ldA;
  for (p=A; p<pend; p+=ldA)  {
    tend = p+inner;
    for (t=p; t<tend; *t++ = value);
  }
}

/* print a dense matrix */
void ArrayPrint(double *A, int ldA, int outer , int inner, FILE *output)
{
  double *p, *pend, *t, *tend;
  
  fprintf(output, "%d %d\n", outer, inner);

  pend = A + outer*ldA;

  for (p=A; p<pend; p+=ldA) {
    tend = p+inner;
    for (t=p; t<tend; t++) 
      fprintf(output,"%f ",*t);
    fprintf(output,"\n");
  }
}

/* save a dense matrix */
void ArraySave(double *A, int ldA, int numRows , int numCols, char
	       *outFileName, enum CBLAS_ORDER order)
{
  double *p, *pend, *t, *tend;
  FILE *outFile;

  if ((outFile = fopen(outFileName, "w")) == NULL) 
    exit(1);

  fprintf(outFile,"%d %d\n", numRows, numCols);
  
  if (order == CblasColMajor) {
    pend = A + numCols*ldA;
    for (p=A; p<pend; p+=ldA) {
      tend = p + numRows;
      for (t=p; t<tend; t++) 
	fprintf(outFile,"%f ",*t);
      fprintf(outFile,"\n");
    }
  }
  else if (order == CblasRowMajor) {
        
    tend = A + numCols;
    for (t = A; t< tend; t++) {
      pend = t + numRows * ldA;
      for (p = t; p < pend; p += ldA)
	fprintf(outFile,"%f ",*p);
      fprintf(outFile,"\n");
    }
  }
  else 
    exit(1);

  fclose(outFile);  
}

/* read in a dense matrix */
void ArrayRead(double *A, int M , int N, FILE *input)
{
  int i;
  int end = M * N;
  
  fscanf(input, "%d %d\n", &M, &N);
  for (i = 0; i < M*N; i++) {
    fscanf(input, "%lg", &A[i]);
  }
}


int Read_mtx_size(FILE *f, int *nb, int *M, int *N, int *nz )
{
    char line[MM_MAX_LINE_LENGTH];
    int num_items_read;

    /* set return null parameter values, in case we exit with errors */
    *M = *N = *nz = 0;

    /* now continue scanning until you reach the end-of-comments */
    do 
    {
        if (fgets(line,MM_MAX_LINE_LENGTH,f) == NULL) 
            return MM_PREMATURE_EOF;
    }while (line[0] == '%');

    /* line[] is either blank or has nb, M,N, nz */
    if (sscanf(line, "%d %d %d", nb, M, N, nz) == 3)
        return 0;
        
    else
    do
    { 
        num_items_read = fscanf(f, "%d %d %d", M, N, nz); 
        if (num_items_read == EOF) return MM_PREMATURE_EOF;
    }
    while (num_items_read != 3);

    return 0;
}

/* convert the DBCO to DCOO and save */
void Save_dbco(char *fname, DBCO *A) {

  int M = A->M;
  int N = A->N;

  int order = A->order_;
  int nz = A->nz_;
  int nb = A->nb_;
  double *val = A->val_;
  int *indi = A->indi_;
  int *indj = A->indj_;
  
  int *I = (int *) malloc(nz * nb * nb* sizeof(int));
  int *J = (int *) malloc(nz * nb * nb *sizeof(int));

  DCOO B;

  B = dbco2coo(order,  nz,  nb, val, indi, indj, I, J);

  mm_write_mtx_crd(fname,  M,  N,  nz,  I, J, val, matcode);
}


int main(int argc, char *argv[])
{
    int ret_code;
    MM_typecode matcode;

    FILE *f, *inB, *inBB, *inBB2, *outC0, *outC1;

    /* values to create the compressed coordinate matrix A */
    int M, N, nb, nz;   
    int i, j, temp, *I, *J, *I1, *J1;
    double *val;
    int properties;
    int transpose = CblasNoTrans;
    int order = CblasColMajor;
    SPBLASI_double_COO A, A1;

    double alpha = 1.0;
    double *B, *BB;
    int ldb = 30;
    int BBN = 60;
    int BBk = 60;
    int ldbb = 2 * ldb;

    /* k == the number of columns of B and C */
    int k = 30;
    
    double beta = 0.0;
    double *C, *CC;
    int ldc = 30;
    int CCM = 60;
    int CCk = 60;
    int ldcc = 2 * ldc;
    char* fname0 = "Data/Acoord";
    char* fname1 = "Data/A1coord";
   

    if ((f = fopen("Data/Block_pores_1_nb5.mtx", "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }
    if ((inB = fopen("Data/B150random", "r")) == NULL) {
      printf("file not found\n");
      exit(1);
    }
    if ((inBB = fopen("Data/BB300random", "r")) == NULL) {
      printf("file not found\n");

      exit(1);
    }
    if (mm_read_banner(f, &matcode) != 0)
    {
      printf("Could not process Matrix Market banner.\n");
      exit(1);
    }




    /*  This is how one can screen matrix types if their application */
    /*  only supports a subset of the Matrix Market data types.      */

    if (mm_is_complex(matcode) && mm_is_matrix(matcode) && 
            mm_is_sparse(matcode) )
    {
      printf("Sorry, this application does not support ");
      printf("Market Market type: [%s]\n", mm_typecode_to_str(matcode));
      exit(1);
    }

    /* find out size of sparse matrix .... */

    if ((ret_code = Read_mtx_size(f, &M, &N, &nb, &nz)) !=0)
        exit(1);


    /* reserve memory for matrix */
    I = (int *) malloc(nz * sizeof(int));
    J = (int *) malloc(nz * sizeof(int));


    /* reserve memory for 1-base indices */
    I1 = (int *) malloc(nz * sizeof(int));
    J1 = (int *) malloc(nz * sizeof(int));
    val = (double *) malloc(nb * nb * nz * sizeof(double));

    /* NOTE: when reading in doubles, ANSI C requires the use of the "l"  */
    /*   specifier as in "%lg", "%lf", "%le", otherwise errors will occur */
    /*  (ANSI C X3.159-1989, Sec. 4.9.6.2, p. 136 lines 13-15)            */

    /* read the compressed coordinate matrix */
    for (i=0; i<nz; i++)
    {
        fscanf(f, "%d %d\n", &I[i], &J[i]);
        /* adjust from 1-based to 0-based */
	
	I1[i] = I[i];
	J1[i] = J[i];

	I[i]--;  
	J[i]--;
    }

    for (i=0; i<nz * nb * nb; i++)
      {
        fscanf(f, "%lg\n", &val[i]);
        /* adjust from 1-based to 0-based */
      }


    if (f !=stdin) fclose(f);


    /************************************
      create the structure A (zero-based)
      ***********************************/
    block_ordering = 1;
    properties = blas_zero_base;
    A = create_DBCO(nb, M, N, nz, val, I, J, properties);


    /************************/
    /* now write out matrix */
    /************************/

    DBCO_print(A);

    /*********************************************
      create the structure A1 (one-based - case 1)
      ********************************************/

    block_ordering = 0;
    properties = blas_one_base;
    A1 = create_DBCO(nb, M, N, nz, val, I1, J1, properties);


    /***********************************
      convert A and A1 to COO and save 
      (for verification by matlab)
      ***********************************/


    Save_dbco(fname0, A);

    Save_dbco(fname1, A1);


    /***********************
      allocate memory for B
     **********************/
    
    B = (double *) malloc(N * k * sizeof(double));
    
    /*************************************************
      read (and print) the matrix B in col-major order
      ************************************************/
    
    ArrayRead(B, N, k, inB);

    /*************
      allocate C
      **************/
    
    C = (double *) malloc(M * k * sizeof(double));

    /*******************************************************
      create and initialize BB, 
      which contains B as a submatrix (column-major - case 12)
      *******************************************************/

    BBN = 2 * N;
    BBk = 2 * k;
    
    BB = (double *) malloc(BBN * BBk * sizeof(double));
    
    /* read the matrix BB in col-major order */

    ArrayRead(BB, BBN , BBk, inBB);
    
    /* create CC, which contains the submatrix C */
    
    CCM = 2 * M;
    CCk = 2 * k;
    
    CC = (double *) malloc(CCM * CCk * sizeof(double));


    /****************************************************************

      Case 0
      transpose = no trans
      alpha = 3.25
      nb = 5
      nb_ordering = col
      beta = -2.75
      ldb = M;
      ldc = N;
      prop = base 0
      order = col

      ****************************************************************/


    /*********************************************
      initialize C in column-major order
      ********************************************/
    
    ArraySetToValue(C,ldc,M,k,0.0);
    
    /********************************************
      multiply the matrix A by B

      C <- A * B
      ********************************************/

    /*    
    
    **** INSERT BLOCK MATRIX MULTIPLY CODE HERE *****


    c_dusmm_CSR(order, transpose, k, alpha, A, B, ldb, beta, C, ldc);
    */

    /********************
      save C to file
      *********************/

    ArraySave(C,ldc,M,k,"Data/outC0",order);

    /*************
      End Case 0
      ************/

 
    /****************************************************************

      Case 1
      transpose = trans
      alpha = 3.25
      nb = 5
      nb_ordering = row
      beta = -2.75
      ldb = 2*M
      ldc = 2*N
      prop = base 1
      order = row

      ****************************************************************/
    
    transpose = CblasTrans;
    order = CblasRowMajor;

    ArraySetToValue(CC, ldcc, CCM, CCk, 0.0);
    
    /*

    **** INSERT BLOCK MATRIX MULTIPLY CODE HERE *****

    c_dusmm_CSR(order, transpose, k, alpha, A1, B, ldb, beta, C, ldc);
    */


    /* just save the nonzero center */
    ArraySave((CC + (ldc * M) + k),ldc,M,k,"Data/outC1",order);

    /*
    ArraySave((CC + (ldc * M) + k),ldcc,CCM,CCk,"Data/outC1",order);
    */

    /************
      END CASE 1
    *************/



    /* free memory */
    free(I);
    free(I1);
    free(J);
    free(J1);
    free(val);
    free(B);
    free(C);
    free(BB);
    free(CC);
    free(BB2);

    destroy_DCOO(A);
    destroy_DCOO(A1);

    return 0;

}
