/*  test_vect_mult.c  (Wed Aug  4 15:47:25 EDT 1999)
 *  Author:  Galen Wilkerson
 *
 *  See README.txt for usage
 *
 *  Significant parts of this code modified from example_read.c
 *  by Roldan Pozo and Karin Remington
 *
 *  Contact:  galen@cam.nist.gov or pozo@cam.nist.gov
 */


#include <stdio.h>
#include <malloc.h>
#include "mmio.h"
#include "spblas_i.h"
#include "c_blas.h"


/* all these functions could be macros */
/* initialize a matrix */
void ArraySetToValue(double *A, int ldA, int outer , int inner, double value)
{
  double *p, *pend, *t, *tend;

  pend = A + outer*ldA;
  for (p=A; p<pend; p+=ldA)  {
    tend = p+inner;
    for (t=p; t<tend; *t++ = value);
  }
}

int main(int argc, char *argv[])
{
    int ret_code;
    MM_typecode matcode;
    FILE *f, *inX, *inX1, *outY, *outY1;
    int M, N, nz;   
    int i, *I, *J, *I1, *J1;
    double *val;
    int properties = sparse_blas_zero_base;
    int transpose = blas_no_trans;

    SPBLASI_double_COO A, A1;
    double alpha = 1.0;
    double *x, *x1;
    int incx = 1;
    int incx1 = 2;

    double beta = 0.0;
    double *y, *y1;
    int incy = 1;
    int incy1 = 1;

    if (argc < 6)
	{
	  fprintf(stderr, "Usage: %s [martix-market-filename] [x-vector filename] [y-vector filename] [x1-vector filename] [y1-vector filename]\n", argv[0]);
	  exit(1);
	}
    else    
    { 
      if ((f = fopen(argv[1], "r")) == NULL) 
	exit(1);
      if ((inX = fopen(argv[2], "r")) == NULL) 
	exit(1);
      if ((outY = fopen(argv[3], "w")) == NULL) 
	exit(1);
      if ((inX1 = fopen(argv[4], "r")) == NULL) 
	exit(1);
      if ((outY1 = fopen(argv[5], "w")) == NULL) 
	exit(1);
    }

    if (mm_read_banner(f, &matcode) != 0)
    {
        printf("Could not process Matrix Market banner.\n");
        exit(1);
    }


    /*  This is how one can screen matrix types if their application */
    /*  only supports a subset of the Matrix Market data types.      */

    if (mm_is_complex(matcode) && mm_is_matrix(matcode) && 
            mm_is_sparse(matcode) )
    {
        printf("Sorry, this application does not support ");
        printf("Market Market type: [%s]\n", mm_typecode_to_str(matcode));
        exit(1);
    }

    /* find out size of sparse matrix .... */

    if ((ret_code = mm_read_mtx_crd_size(f, &M, &N, &nz)) !=0)
        exit(1);


    /* reserve memory for matrices */

    I = (int *) malloc(nz * sizeof(int));
    J = (int *) malloc(nz * sizeof(int));
    I1 = (int *) malloc(nz * sizeof(int));
    J1 = (int *) malloc(nz * sizeof(int));
    val = (double *) malloc(nz * sizeof(double));


    /* NOTE: when reading in doubles, ANSI C requires the use of the "l"  */
    /*   specifier as in "%lg", "%lf", "%le", otherwise errors will occur */
    /*  (ANSI C X3.159-1989, Sec. 4.9.6.2, p. 136 lines 13-15)            */

    /* read the compressed coordinate matrix */
    for (i=0; i<nz; i++)
    {
        fscanf(f, "%d %d %lg\n", &I1[i], &J1[i], &val[i]);
        /* adjust from 1-based to 0-based */
	I[i] = I1[i] - 1;  
	J[i] = J1[i] - 1;
    }

    if (f !=stdin) fclose(f);

    /************************/
    /* now write out matrix */
    /************************/

    mm_write_banner(stdout, matcode);
    mm_write_mtx_crd_size(stdout, M, N, nz);
    for (i=0; i<nz; i++)
      fprintf(stdout, "%d %d %20.19g\n", I[i]+1, J[i]+1, val[i]);
    
    /************************
      create the structure A
      read the vector x
      create the vector y
    ************************/
    
    A = create_DCOO(M, N, nz, val, I, J, properties);
    printf("\nDCOO A:\n");
    DCOO_print(A);

    properties = sparse_blas_one_base;
    A1 = create_DCOO(M, N, nz, val, I1, J1, properties);
    printf("\nDCOO A1:\n");
    DCOO_print(A1);
 
    x = (double *) malloc(incx * N * sizeof(double));
    x1 = (double *) malloc(incx1 * N * sizeof(double));

 

    /**********************
      read the x and x1-vector
      **********************/
    
    printf("\n x= \n");
    for (i = 0; i < ( incx * N ); i++) {
      fscanf(inX, "%lg\n", &x[i]);
      fprintf(stdout, "%20.15g\n", x[i]);
    }

    printf("\n x1= \n");
    for (i = 0; i < ( incx1 * N ); i++) {
      fscanf(inX1, "%lg\n", &x1[i]);
      fprintf(stdout, "%20.15g\n", x1[i]);
    }
    
    y = (double *) malloc(incy * M * sizeof(double));
    y1 = (double *) malloc(incy1 * M * sizeof(double));


    /****************************
      Case 0
      transpose = no trans
      incx = 1
      incy = 1
      prop = base 0

      ****************************/

    
    /******************
      initialize y
      ******************/

    ArraySetToValue(y, M, 1 ,M, 0.0);



    /********************************************
      multiply the matrix A by x, a vector of 1's
      y = A*x
      
      ********************************************/
    
    c_dusmv_COO(transpose, alpha, A, x, incx, beta, y, incy);


    /********************
      print the vector y
      and save it to file
      *********************/
    
    printf("\ny = \n");
     
    for (i=0; i< (incy * M); i++) {
      /*
	fprintf(stdout, "%20.15g\n", *(y+i));
	*/
      fprintf(outY, "%20.15g\n", *(y+i));
    }


    /****************************
      Case 1
      transpose = trans
      incx1 = 2
      incy1 = 1
      prop = base 1

      ****************************/
    
    transpose = blas_trans;

     /******************
      initialize y1
      ******************/
    
    
    ArraySetToValue(y1, M, 1 ,M, 0.0);

     
    c_dusmv_COO(transpose, alpha, A1, x1, incx1, beta, y1, incy1);


    /********************
      print the vector y1
      and save it to file
      *********************/
    
    printf("\ny1 = \n");
     
    for (i=0; i< (incy1 * M); i++) {
      /*
	fprintf(stdout, "%20.15g\n", *(y1+i));
	*/
      fprintf(outY1, "%20.15g\n", *(y1+i));
    }




    /* free memory */
    free(I);
    free(I1);
    free(J);
    free(J1);
    free(val);
    free(x);
    free(x1);
    free(y);
    free(y1);

    destroy_DCOO(A);
    destroy_DCOO(A1);

    return 0;
}
