/* 
*   Matrix Market I/O example program
*
*   Read a real (non-complex) sparse matrix from a Matrix Market (v. 2.0) file.
*   and copies it to stdout.  This porgram does nothing useful, but
*   illustrates common usage of the Matrix Matrix I/O routines.
*   (See http://math.nist.gov/MatrixMarket for details.)
*
*   Usage:  a.out [filename] > output
*
*       
*   NOTES:
*
*   1) Matrix Market files are always 1-based, i.e. the index of the first
*      element of a matrix is (1,1), not (0,0) as in C.  ADJUST THESE
*      OFFSETS ACCORDINGLY offsets accordingly when reading and writing 
*      to files.
*
*   2) ANSI C requires one to use the "l" format modifier when reading
*      double precision floating point numbers in scanf() and
*      its variants.  For example, use "%lf", "%lg", or "%le"
*      when reading doubles, otherwise errors will occur.
*/

#include <stdio.h>
#include <malloc.h>
#include "mmio.h"
#include "coomv.h"
#include "spblas_i.h"

int main(int argc, char *argv[])
{
    int ret_code;
    MM_typecode matcode;
    FILE *f, *inX, *outY;
    int M, N, nz;   
    int i, *I, *J;
    double *val;
    int properties = blas_zero_base;
    int transpose = blas_no_trans;

    SPBLASI_double_COO A;
    double alpha = 1.0;
    double *x;
    int incx = 2;

    double beta = 0.0;
    double *y, *y2;
    int incy = 3;

    if (argc < 4)
	{
	  fprintf(stderr, "Usage: %s [martix-market-filename] [x-vector filename] [y-vector filename]\n", argv[0]);
	  exit(1);
	}
    else    
    { 
      if ((f = fopen(argv[1], "r")) == NULL) 
	exit(1);
      if ((inX = fopen(argv[2], "r")) == NULL) 
	exit(1);
      if ((outY = fopen(argv[3], "w")) == NULL) 
	exit(1);
    }

    if (mm_read_banner(f, &matcode) != 0)
    {
        printf("Could not process Matrix Market banner.\n");
        exit(1);
    }


    /*  This is how one can screen matrix types if their application */
    /*  only supports a subset of the Matrix Market data types.      */

    if (mm_is_complex(matcode) && mm_is_matrix(matcode) && 
            mm_is_sparse(matcode) )
    {
        printf("Sorry, this application does not support ");
        printf("Market Market type: [%s]\n", mm_typecode_to_str(matcode));
        exit(1);
    }

    /* find out size of sparse matrix .... */

    if ((ret_code = mm_read_mtx_crd_size(f, &M, &N, &nz)) !=0)
        exit(1);


    /* reserve memory for matrices */

    I = (int *) malloc(nz * sizeof(int));
    J = (int *) malloc(nz * sizeof(int));
    val = (double *) malloc(nz * sizeof(double));


    /* NOTE: when reading in doubles, ANSI C requires the use of the "l"  */
    /*   specifier as in "%lg", "%lf", "%le", otherwise errors will occur */
    /*  (ANSI C X3.159-1989, Sec. 4.9.6.2, p. 136 lines 13-15)            */

    /* read the compressed coordinate matrix */
    for (i=0; i<nz; i++)
    {
        fscanf(f, "%d %d %lg\n", &I[i], &J[i], &val[i]);
        /* adjust from 1-based to 0-based */
	I[i]--;  
	J[i]--;
    }

    if (f !=stdin) fclose(f);

    /************************/
    /* now write out matrix */
    /************************/

    mm_write_banner(stdout, matcode);
    mm_write_mtx_crd_size(stdout, M, N, nz);
    for (i=0; i<nz; i++)
      fprintf(stdout, "%d %d %20.19g\n", I[i]+1, J[i]+1, val[i]);
    
    /************************
      create the structure A
      read the vector x
      create the vector y
    ************************/
    
    A = create_DCOO(M, N, nz, val, I, J, properties);
 
    x = (double *) malloc(incx * N * sizeof(double));
 

    /**********************
      read the x-vector
      **********************/
    
    printf("\n x= \n");
    for (i = 0; i < ( incx * N ); i++) {
      fscanf(inX, "%lg\n", &x[i]);
      
      fprintf(stdout, "%f\n", x[i]);
    }

    
    /******************
      initialize y
      ******************/
    
    y = (double *) malloc(incy * M * sizeof(double));

    for (i = 0; i < ( incy * M ); i++) {
      y[i] = 0.0;
    }
    
    /********************************************
      multiply the matrix A by x, a vector of 1's
      y = A*x
      
      ********************************************/
    
    c_dusmv_COO(transpose, alpha, A, x, incx, beta, y, incy);


    /********************
      print the vector y
      and save it to file
      *********************/
    
    printf("\ny = \n");
     
    for (i=0; i< (incy * M); i++) {
      fprintf(stdout, "%f\n", *(y+i));
      fprintf(outY, "%f\n", *(y+i));
    }

    /****************************
      
      y2 = [0,1,2,3,...]
      
      y2 = alpha*A*x + beta*y2
      
      ****************************/
    /*
    beta = 1.0;
    y2 = (double *) malloc(M * sizeof(double));

    printf("initializing y2:\n");

    for (i = 0; i < M; i++) {
      y2[i] = (double) i;
      fprintf(stdout, "%f\n", *(y2+i));
    }

    c_dusmv_COO(transpose, alpha, A, x, incx, beta, y2, incy);

    printf("\ny2 = \n");
    
    for (i=0; i< M; i++)
      fprintf(stdout, "%f\n", *(y2+i));
      */
    return 0;
}


