#include "c_blas.h"

/* converter to BCOO to COO:

	BCOO is a block-structed sparse matrix, with block i
	starting at location (indi[i], indj[i]) and having size
	nb*nb.

	Basically creates a new I, J index vector of length nz*nb*nb.
	Note that the val[] is not really used.

	In the function below, val[] is of length nz*nb*nb, 
	indi[] & indj[] are of length nz, and I&J are of length
	nz*nb*nb.

	NOTE: do not need to worry about 0/1 bases, since we
	are not changing convention between the block and point-entry
	representations.

*/
void dbco2coo(enum blas_order_type order, int nz, int nb, 
		const double *val, const int *indi, const int *indj, int *I, int *J)
{

	int i, j, k;
	int p;


	if (order == blas_rowmajor)
	{
		for (k=0; k<nz; k++)		/* for each block */
		{
			p = k * nb*nb;
			for (i=0; i<nb; i++)
			{
				for (j=0; j<nb; j++)
				{
					I[p] = indi[k]+i;
					J[p] = indj[k]+j;
					p++;
				}
			}
		}
	}
	else
	{
		for (k=0; k<nz; k++)		/* for each block */
		{
			p = k * nb*nb;
			for (j=0; j<nb; j++)
			{
				for (i=0; i<nb; i++)
				{
					I[p] = indi[k]+i;
					J[p] = indj[k]+j;
					p++;
				}
			}
		}
	}
}

#ifdef TEST_BCO2COO

#include <stdio.h>
int main()
{
	int i=0;

#define nz 6
#define nb 3
#define nznbnb 54

	int indi[] = {  10,  20,  30,  40,  50,  60};
	int indj[] = { 100, 200, 300, 400, 500, 600};

	int I[nznbnb] = { 0 };
	int J[nznbnb] = { 0 };

	dbco2coo(blas_rowmajor, nz, nb, NULL, indi, indj, I, J);

	printf("row order: \n");
	for (i=0; i<nz*nb*nb; i++)
		printf("%4d %4d\n", I[i], J[i]);


	dbco2coo(blas_colmajor, nz, nb, NULL, indi, indj, I, J);
	printf("\n\ncol order: \n");
	for (i=0; i<nz*nb*nb; i++)
		printf("%4d %4d\n", I[i], J[i]);


	return 0;
}
	
#endif

