/*
 *                 Author:  Christopher G. Phillips
 *              Copyright (C) 1993 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * The author makes no representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#ifdef USE_GDBM
#include <gdbm.h>
#define dbm_delete			gdbm_delete
#define dbm_fetch			gdbm_fetch
#define dbm_firstkey			gdbm_firstkey
#define dbm_nextkey(dbmfile)		gdbm_nextkey(dbmfile, key)
#define dbm_open(file, perm, mode)	gdbm_open(file, 0, perm, mode, NULL)
#define dbm_store			gdbm_store
#define DBM_INSERT			GDBM_INSERT
#define DBM_REPLACE			GDBM_REPLACE
typedef GDBM_FILE	dbmfile_t;
#else
#include <ndbm.h>
typedef DBM		*dbmfile_t;
#endif
#include "leak.h"

#undef malloc
#undef realloc
#undef free
#undef calloc
#undef strdup

#define FILE_DBMFILENAME	"filedbm"
#define MEMORY_DBMFILENAME	"memdbm"

int	leak_logging = 1;

typedef char	*DPTR;

static dbmfile_t	filedbm = NULL;
static dbmfile_t	memdbm = NULL;
static int		nextfile = -1;
static int		dbms_zapped = 0;

struct memdata {
	size_t	size;
	int	file;
	int	line;
};

static void
opendbmfile(dbmfile_t *dbmp, /*const*/ char *file)
{
	if (*dbmp == NULL && (*dbmp = dbm_open(file, O_RDWR | O_CREAT,
	  S_IRUSR | S_IWUSR)) == NULL) {
		perror("dbm_open");
		exit(1);
	}
}

static int
file2int(const char *file)
{
	datum	key, data;

	key.dptr = (DPTR)file;
	key.dsize = strlen(file) + 1;
	data = dbm_fetch(filedbm, key);
	if (data.dptr) {
		int	i;

		(void)memcpy(&i, data.dptr, sizeof i);
		return i;
	} else {
		++nextfile;
		data.dptr = (DPTR)&nextfile;
		data.dsize = sizeof nextfile;
		if (dbm_store(filedbm, key, data, DBM_INSERT) < 0) {
			perror("dbm_store");
			exit(1);
		}
		return nextfile;
	}
}

static char *
int2file(int file)
{
	datum	key, data;
	int	i;

	for (key = dbm_firstkey(filedbm); key.dptr;
	  key = dbm_nextkey(filedbm)) {
		data = dbm_fetch(filedbm, key);
		if (data.dptr == NULL) {
			perror("dbm_fetch");
			continue;
		}
		(void)memcpy(&i, data.dptr, sizeof i);
		if (i == file)
			return key.dptr;
	}
	return "???";
}

static void
leak_fclear(void)
{
	if (filedbm)
		(void)dbm_close(filedbm);
#ifdef USE_GDBM
	(void)remove(FILE_DBMFILENAME);
#else
	(void)remove(FILE_DBMFILENAME ".dir");
	(void)remove(FILE_DBMFILENAME ".pag");
#endif
	filedbm = NULL;
}

void
leak_clear(void)
{
	if (memdbm)
		(void)dbm_close(memdbm);
#ifdef USE_GDBM
	(void)remove(MEMORY_DBMFILENAME);
#else
	(void)remove(MEMORY_DBMFILENAME ".dir");
	(void)remove(MEMORY_DBMFILENAME ".pag");
#endif
	memdbm = NULL;
}

void
leak_insert(const void *p, size_t size, const char *file, int line, int mode)
{
	struct memdata	md;
	datum		key, data;

	if (!dbms_zapped) {
		leak_clear();
		leak_fclear();
		dbms_zapped = 1;
	}
		
	if (!filedbm)
		opendbmfile(&filedbm, FILE_DBMFILENAME);
	if (!memdbm)
		opendbmfile(&memdbm, MEMORY_DBMFILENAME);

	md.file = file2int(file);

	md.size = size;
	md.line = line;
	key.dptr = (DPTR)&p;
	key.dsize = sizeof p;
	data.dptr = (DPTR)&md;
	data.dsize = sizeof md;

	if (dbm_store(memdbm, key, data, mode) == -1) {
		perror("dbm_store");
		exit(1);
	}
}

void
leak_delete(const void *p, const char *file, int line)
{
	datum	key;

	if (!dbms_zapped) {
		leak_clear();
		leak_fclear();
		dbms_zapped = 1;
	}

	key.dptr = (DPTR)&p;
	key.dsize = sizeof p;

	if (!memdbm)
		opendbmfile(&memdbm, MEMORY_DBMFILENAME);
	if (dbm_delete(memdbm, key) == -1) {
		fprintf(stderr,
		  "free unmalloced pointer %p from \"%s\", line %d\n", p, file,
		  line);
	}
}

void
leak_dump(void)
{
	struct memdata	md;
	datum		key, data;
	void		*addr;
	int		do_title = 1;

	if (!filedbm)
		opendbmfile(&filedbm, FILE_DBMFILENAME);
	if (!memdbm)
		opendbmfile(&memdbm, MEMORY_DBMFILENAME);

	for (key = dbm_firstkey(memdbm); key.dptr; key = dbm_nextkey(memdbm)) {
		data = dbm_fetch(memdbm, key);
		if (data.dptr == NULL) {
			perror("dbm_fetch");
			continue;
		}
		if (do_title) {
			int	i;

			printf("Nonfreed blocks:\n");
			printf("Address \tSize\tLine\tFile\n");
			for (i = 0; i < 36; i++)
				putchar('-');
			putchar('\n');
			do_title = 0;
		}
		/*
		 * Copy the key and data
		 * because they may be overwritten in int2file.
		 */
		md = *(struct memdata *)data.dptr;
		addr = *(void **)key.dptr;
#ifndef sun
		printf("%8p\t%d\t%d\t%s\n", addr,
#else
		printf("%08x\t%d\t%d\t%s\n", (int)addr,
#endif
		  md.size, md.line, int2file(md.file));
	}
}

void *
leak_malloc(size_t size, const char *file, int line)
{
	void	*ptr = malloc(size);

	if (leak_logging && ptr)
		leak_insert(ptr, size, file, line, DBM_INSERT);

	return ptr;
}

void *
leak_realloc(void *oldptr, size_t size, const char *file, int line)
{
	void	*newptr = realloc(oldptr, size);

	if (leak_logging) {
		if (!size && oldptr)
			leak_delete(oldptr, file, line);
		else if (newptr) {
			if (newptr == oldptr)
				leak_insert(newptr, size, file, line,
				  DBM_REPLACE);
			else {
				if (oldptr)
					leak_delete(oldptr, file, line);
				leak_insert(newptr, size, file, line,
				  DBM_INSERT);
			}
		}
	}

	return newptr;
}

void
leak_free(void *ptr, const char *file, int line)
{
	free(ptr);

	if (leak_logging && ptr)
		leak_delete(ptr, file, line);
}

void *
leak_calloc(size_t s, size_t t, const char *file, int line)
{
	void	*ptr = calloc(s, t);

	if (leak_logging && ptr)
		leak_insert(ptr, s * t, file, line, DBM_INSERT);

	return ptr;
}

char *
leak_strdup(const char *s, const char *file, int line)
{
	size_t	size = strlen(s) + 1;
#ifdef HAS_STRDUP
	char	*ptr = strdup(s);
#else
	char	*ptr = malloc(size);

	if (ptr)
		strcpy(ptr, s);
#endif

	if (leak_logging && ptr)
		leak_insert(ptr, size, file, line, DBM_INSERT);

	return ptr;
}
