#include <stdio.h>
#include <stdlib.h>
#include "c_blas.h"
#include "microblas.h"


/* -------------------  GENERALIZED SPARSE DDOT ------------------------ */
/*                                                                       */
/*                 r = alpha * dot(x,y) + beta * r;                      */
/*                                                                       */
/* ----------------------------------------------------------------------*/

/* r = alpha * dot(x, y)  + beta * r */
void ublas_spdotab(double alpha, int ibegin, int iend, 
		const double *x, const int *ind, const double *y, int incy,
		double beta, double *r)
{ 
	int i; 
	double t= 0.0;
	for (i=ibegin; i<iend; i++) 
		t += x[i] * y[ind[i]*incy]; 
	*r = alpha * t + beta * *r;
}

/* ---------------------  GENERALIZED DAXPBY --------------------------- */
/*                                                                       */
/*                   y = alpha * x  + beta * y                           */ 
/*                                                                       */
/* ----------------------------------------------------------------------*/


/* y = alpha * x  + beta*y */
void ublas_spdaxpby(int N, int ibegin, int iend, double alpha, 
	const double *x, const int *ind, double beta, double *y, int incy) 
{ 
	int i; 
	int Nend = N*incy; 
 	for (i=0; i<Nend; i+= incy) 
		y[i] *= beta; 
	 
	for (i=ibegin; i<iend; i++) 
	{
		y[ind[i] * incy] += alpha * x[i]; 
	}
}


/* ---------------- GENERALIZED BLOCK SPARSE DOT PRODUCT ----------------*/
/*                                                                       */
/*                   C =  alpha * A * B + beta * C                       */
/*                                                                       */
/* ----------------------------------------------------------------------*/

void ublas_bspdotab_row(int nb, enum blas_trans_type blocktransA, 
	int Mb, int k, double alpha,  int ibegin, int iend, 
	const double *val, const int *indx, 
	const double *B, int ldB, double beta, double *C, int ldC)
{
	const int nb2 = nb*nb;
	int i=0;

	if (!(ibegin < iend)) return;

	/* first time through: C = alpha A1 * B + beta * C */
	c_dgemm(blas_rowmajor, blocktransA, blas_no_trans, nb, nb, k, alpha,
			&val[ibegin*nb2], nb, &B[indx[ibegin]*ldB*nb], ldB, beta,
			C, ldC);

	/* subsequent times:   C = alpha Ai * B + C    (alpha=1.0) */
	for (i=ibegin+1; i<iend; i++)
	{
		c_dgemm(blas_rowmajor, blocktransA, blas_no_trans, nb, nb, k, alpha,
			&val[i*nb2], nb, &B[indx[i]*ldB*nb], ldB, 1.0,
			C, ldC);
	}
}

void ublas_bspdotab_col(
	int nb, enum blas_trans_type blocktransA, 
	int Mb, int k, double alpha, 
	int ibegin, int iend, const double *val, const int *indx, 
	const double *B, int ldB, 
	double beta, 
	double *C, int ldC)
{
	int i=0;
	const int nb2 = nb*nb;

	if (!(ibegin < iend)) return;

	/* first time through: C = alpha A1 * B + beta * C */
	c_dgemm(blas_colmajor, blocktransA, blas_no_trans, nb, nb, k, alpha,
			&val[ibegin*nb2], nb, &B[indx[ibegin]*nb], ldB, beta,
			C, ldC);

	/* subsequent times:   C = alpha Ai * B + C    (alpha=1.0) */
	for (i=ibegin+1; i<iend; i++)
	{
		c_dgemm(blas_colmajor, blocktransA, blas_no_trans, nb, nb, k, alpha, 
		&val[i*nb2], nb, &B[indx[i]*nb], ldB, 1.0, 
		C, ldC);
	}
}

