#include <stdio.h>
#include <malloc.h>
#include "spblas_i.h"
#include "c_blas.h"

DCOO create_DCOO( int M, int N, int nz, double *val,
    int *indx, int *jndx, int properties)

{
    DCOO A = (DCOO) malloc(sizeof(DCOO_struct));

    if (A == NULL) return NULL;

    A->val_ = val;
    A->i_ = indx;
    A->j_ = jndx;
    A->M_ = M;
    A->N_ = N;
    A->nz_ = nz;
    A->prop_ = properties;
    A->reclaim_space = 0;

    return A;
}

DCOO create_DCOO_copy( int M, int N, int nz, double *val,
    int *indx, int *jndx, int properties)

{
    DCOO A = NULL;
    int i;

    if ((A = (DCOO) malloc(sizeof(DCOO_struct))) == NULL) return NULL;
    if ((A->val_ = (double*) malloc(sizeof(double)*nz)) == NULL) return NULL;
    if ((A->i_ = (int*) malloc(sizeof(int)*nz)) == NULL) return NULL;
    if ((A->j_ = (int*) malloc(sizeof(int)*nz)) == NULL) return NULL;

    /* copy over values */

    for (i=0; i<nz; i++)
    {
        A->val_[i] = val[i];
        A->i_[i] = indx[i];
        A->j_[i] = jndx[i];
    }
    A->M_ = M;
    A->N_ = N;
    A->nz_ = nz;
    A->prop_ = properties;
    A->reclaim_space = 1;

    return A;
}


void destroy_DCOO( DCOO A)
{
    if (A==NULL) return;

    if (A->reclaim_space == 1)
    {
        free(A->val_);
        free(A->i_);
        free(A->j_);
    }
    
    free(A);
}


void DCOO_print(DCOO A)
{
	int i=0;
	int nz = A->nz_;
	double *val = A->val_;
	int *I = A->i_;
	int *J = A->j_;

	for (i=0; i< nz; i++)
		printf("%10.4g   %10d %10d \n", val[i], I[i], J[i]);
}
	
	
DCSR create_DCSR( int M, int N, int nz, double *val,
    int *indx, int *pntrb, int *pntre, int properties)

{
    DCSR A = (DCSR) malloc(sizeof(DCSR_struct));

    if (A == NULL) return NULL;

    A->val_ = val;
    A->indx_ = indx;
    A->pntre_ = pntre;
    A->pntrb_ = pntrb;
    A->M_ = M;
    A->N_ = N;
    A->nz_ = nz;
    A->prop_ = properties;
    A->reclaim_space = 0;

    return A;
}

/* note: we could be even more frugal with storage, in the case
    when pntrb and pntre overlapp the same vector.  But it is not
    terribly critical; also, it make the freeing (destroying) a
    little more complicated, as one cannot free pntrb[] and
    pntre[] separately.
*/
DCSR create_DCSR_copy( int M, int N, int nz, double *val,
    int *indx, int *pntrb, int *pntre, int properties)

{
    DCSR A = NULL;
    int i;

    if ((A = (DCSR) malloc(sizeof(DCSR_struct))) == NULL) return NULL;
    if ((A->val_ = (double*) malloc(sizeof(double)*nz)) == NULL) return NULL;
    if ((A->indx_ = (int*) malloc(sizeof(int)*nz)) == NULL) return NULL;
    if ((A->pntrb_ = (int*) malloc(sizeof(int)*M)) == NULL) return NULL;
    if ((A->pntre_ = (int*) malloc(sizeof(int)*M)) == NULL) return NULL;

    /* copy over values */

    for (i=0; i<nz; i++)
    {
        A->val_[i] = val[i];
        A->indx_[i] = indx[i];
    }

    for (i=0; i<M; i++)
    {
        A->pntrb_[i] = pntrb[i];
        A->pntre_[i] = pntre[i];
    }
    A->M_ = M;
    A->N_ = N;
    A->nz_ = nz;
    A->prop_ = properties;
    A->reclaim_space = 1;

    return A;
}


void destroy_DCSR( DCSR A)
{
    if (A==NULL) return;

    if (A->reclaim_space == 1)
    {
        free(A->val_);
        free(A->indx_);
        free(A->pntrb_);
        free(A->pntre_);
    }
    
    free(A);
}

void DCSR_print(DCSR A)
{
	int row=0;
	int i=0;
	double *val = A->val_;
	int *col = A->indx_;
	int *pntrb = A->pntrb_;
	int *pntre = A->pntre_;
	int index_base = A->prop_ || sparse_blas_zero_base;

	for (row=0; row< A->M_; row++)
		for (i=(pntrb[row]-index_base); i<(pntre[row]-index_base); i++)
			printf(" %10.4g  %10d  %10d\n", val[i], row + index_base, col[i]);
}

void DCSR_print_i(DCSR A)
{
	int row=0;
	int i=0;
	int nz = A->nz_;
	double *val = A->val_;
	int *col = A->indx_;
	int *pntrb = A->pntrb_;
	int *pntre = A->pntre_;

	/* print out row pointer */
	printf(" pntrb[]   pntre[]: \n");
	for (row=0; row< A->M_; row++)
		printf("%10d  %10d\n", pntrb[row], pntre[row]);

	/* print out nonzeros */
	printf(" nonzeros: \n");
	for (i=0; i<nz; i++)
		printf("%10.4g\n", val[i]);

	/* print out column indices  */
	printf(" column indices: \n");
	for (i=0; i<nz; i++)
		printf("%10d\n", col[i]);

}
	



/* compressed - column */
	
DCSC create_DCSC( int M, int N, int nz, double *val,
    int *indx, int *pntrb, int *pntre, int properties)

{
    DCSC A = (DCSC) malloc(sizeof(DCSC_struct));

    if (A == NULL) return NULL;

    A->val_ = val;
    A->indx_ = indx;
    A->pntre_ = pntre;
    A->pntrb_ = pntrb;
    A->M_ = M;
    A->N_ = N;
    A->nz_ = nz;
    A->prop_ = properties;
    A->reclaim_space = 0;

    return A;
}


void destroy_DCSC( DCSC A)
{
    if (A==NULL) return;

    if (A->reclaim_space == 1)
    {
        free(A->val_);
        free(A->indx_);
        free(A->pntrb_);
        free(A->pntre_);
    }
    
    free(A);
}

void DCSC_print(DCSC A)
{
	int col=0;
	int i=0;
	double *val = A->val_;
	int *row = A->indx_;
	int *pntrb = A->pntrb_;
	int *pntre = A->pntre_;
	int index_base = A->prop_ || sparse_blas_zero_base;

	for (col=0; col< A->N_; col++)
		for (i=(pntrb[col]-index_base); i<(pntre[col]-index_base); i++)
			printf(" %10.4g  %10d  %10d\n", val[i], row[i], col + index_base);
}

void DCSC_print_i(DCSC A)
{
	int col=0;
	int i=0;
	int nz = A->nz_;
	double *val = A->val_;
	int *row = A->indx_;
	int *pntrb = A->pntrb_;
	int *pntre = A->pntre_;

	/* print out col pointer */
	printf(" pntrb[]   pntre[]: \n");
	for (col=0; col< A->M_; col++)
		printf("%10d  %10d\n", pntrb[col], pntre[col]);

	/* print out nonzeros */
	printf(" nonzeros: \n");
	for (i=0; i<nz; i++)
		printf("%10.4g\n", val[i]);

	/* print out row indices  */
	printf(" row indices: \n");
	for (i=0; i<nz; i++)
		printf("%10d\n", row[i]);

}
	
	
