/**

  DCSR.java - double-precision Compressed Row Format

  @author Galen J. Wilkerson

 */

import java.io.*;

public class DCSR {
  
  int M_, N_, nz_;
  int prop_;
  double[] val_;          /* length = nz */
  int[] indx_;             
  int[] pntrb_;             
  int[] pntre_;             

  /** 
    constructor
   */
  public DCSR( final int M, final int N, final int nz, final double[] val, final int[] indx, 
	       final int[] pntrb, final int[] pntre, final int properties) 
  {
    M_ = M;
    N_ = N;
    nz_ = nz;

    val_ = val;
    indx_ = indx;
    pntrb_ = pntrb;
    pntre_ = pntre;

    prop_ = properties;
  }

  /**
    conversion constructor - converts from DCOO -> DCSR
   */
  public DCSR (DCOO A) {
    int tempi=0;
    int tempM = A.M_;
    int tempN = A.N_;
    int tempI[] = A.indx_;
    int tempJ[] = A.jndx_;
    double tempval[] = A.val_;
    int tempnz = A.nz_;
    int tempproperties = A.prop_;
    int tempindex = 0;
   
    /* components for CSR routine */
    double csr_val[] = new double[tempnz];
    int csr_pntrb[] = new int[tempM];
    int csr_pntre[] = new int[tempM];
    int csr_j[] = new int[tempnz];
    int tally[] = new int[tempM];
    
    /* 1st pass: find out how many elements in each row */

    for (int i=0; i<tempnz; i++)
      tally[tempI[i]]++;

    /* convert tally[] into (0-based) row pointer by accumulating
       row-lengths.  tally[i] now points to the begining or row i  */
    
    for (int i = 1;i < tempM; i++)
      tally[i] += tally[i-1];

    /* copy these values into pntrb[]; (we will be using the values
       of tally[] as place-markers, which change as we fill up the
       CSR matrix). */

    csr_pntrb[0]=0;
    for (int i=0; i < tempM-1; i++) {
      csr_pntrb[i+1] = tally[i];
      csr_pntre[i] = tally[i];
    }
    csr_pntre[tempM-1] = tally[tempM-1];

    /* copy back original row-starts into tally */
    for (int i=0; i < tempM; i++)
      tally[i] = csr_pntrb[i];

    /* 2nd pass: now go through and fill-in CSR format */
    for (int i=0; i < tempnz; i++)
      {
	tempindex = tally[tempI[i]]++;
	csr_val[tempindex] = tempval[i];
	csr_j[tempindex] = tempJ[i];
      }

    //    return create_DCSR(M, N, nz, csr_val, csr_j, csr_pntrb, csr_pntrb+1, 
    //	       properties);

    M_ = tempM;
    N_ = tempN;
    nz_ = tempnz;

    val_ = csr_val;
    indx_ = csr_j;
    pntrb_ = csr_pntrb;
    pntre_ = csr_pntre;

    prop_ = tempproperties;
    return;
    
  }


  public int getNumRows() {
    return M_;
  }

  public int getNumCols() {
    return N_;
  }

  public int getNumNonZero() {
    return nz_;
  }

  /**
    print the current object
   */
  public void print() {
    System.out.println("M " + M_ + " N " + N_ + " nz " + nz_ );

    System.out.println("\nval");
    for (int i = 0; i < val_.length; i++) {
      System.out.println(val_[i] + " ");
    }

    System.out.println("\nindx");
    for (int i = 0; i < val_.length; i++) {
      System.out.println(indx_[i] + " ");
    }

    System.out.println("\npntrb");
    for (int i = 0; i < pntrb_.length; i++) {
      System.out.println(pntrb_[i] + " ");
    }

    System.out.println("\npntre");
    for (int i = 0; i < pntre_.length; i++) {
      System.out.println(pntre_[i] + " ");
    }

    System.out.println("\nproperties " + prop_);
  }

  public void print2() {

    System.out.println("M " + M_ + " N " + N_ + " nz " + nz_);

    for (int tempRow = 0; tempRow < M_; tempRow++) 
      for (int i = pntrb_[tempRow]; i < pntre_[tempRow]; i++)
	System.out.println(val_[i] + "       " + tempRow + "       " + indx_[i]);

  }

  /**
    method to support CaABbC matrix-matrix multiplication.
   */
  public void dusmm(final int transpose, final int k, final double alpha,
			  final double B[][], final double beta, double C[][]) 
  { 

    double t;
    int jb, je;

    if (alpha == 0) 
      {
	if (beta != 1.0) {
	  Dense.arrayMultScalar(C, beta); 
	  return;
	}
      }
    
    // handle transpose with DCSC
    if (transpose == Jblas.Trans) { 

      // need to avoid infinite recursion
      int tempTrans = Jblas.NoTrans;

      // create Compressed Column object
      DCSC dcsc1 = new DCSC(M_, N_, nz_, val_, indx_, pntrb_, pntre_, prop_);

      // multiply
      dcsc1.dusmm(tempTrans, k, alpha, B, beta, C);

      return;
    }

    
    if (alpha == 1.0) {
      for (int p = 0; p < k; p++) {
	for (int i = 0; i < M_; i++) {
	  
	  if (beta == 0.0) 
	    t = 0.0;
	  else if (beta == 1.0)
	    t = C[i][p];
	  else /* beta != 1.0 */
	    t = beta * C[i][p];
	  
	  jb = pntrb_[i];
	  je = pntre_[i];

	  for (int j = jb; j < je; j++) 
	    t += B[indx_[j]][p] * val_[j];
	  C[i][p] = t;
	}
      }
    }
    else {  // alpha != 1.0
      for (int p = 0; p < k; p++) {
	for (int i = 0; i < M_; i++) {
	  if (beta == 0.0) 
	    t = 0.0;
	  else if (beta == 1.0)
	    t = C[i][p];
	  else /* beta != 1.0 */
	    t = beta * C[i][p];
	  
	  jb = pntrb_[i];
	  je = pntre_[i];

	  for (int j = jb; j < je; j++) 
	    t += alpha * B[indx_[j]][p] * val_[j];
	  C[i][p] = t;
	}
      }
    }
  }


  /**
    method to support CaABbC matrix-matrix multiplication using submatrices
    */
  public void dusmm(final int transpose, final int k, final double alpha, final double B[][], 
			  final int Bi, final int Bj, final double beta, double C[][], final int Ci, final int Cj)
  { 

    double t;
    int jb, je;


    if (alpha == 0) 
      {
	if (beta != 1.0) {
	  Dense.arrayMultScalar(C, beta); 
	  return;
	}
      }

    // handle this with DCSC 
    if (transpose == Jblas.Trans) {
      
      // need to avoid infinite recursion
      int tempTrans = Jblas.NoTrans;
      
      // create Compressed Column object
      DCSC dcsc1 = new DCSC(M_, N_, nz_, val_, indx_, pntrb_, pntre_, prop_);
      
      // multiply
      dcsc1.dusmm(tempTrans, k, alpha, B, Bi, Bj, beta, C, Ci, Cj);
      
      return;
    }
    
    if (alpha == 1.0) {
      for (int p = 0; p < k; p++) {
	for (int i = 0; i < M_; i++) {
	  if (beta == 0.0) 
	    t = 0.0;
	  else if (beta == 1.0)
	    t = C[i + Ci][p + Cj];
	  else /* beta != 1.0 */
	    t = beta * C[i + Ci][p + Cj];
	  
	  jb = pntrb_[i];
	  je = pntre_[i];

	  for (int j = jb; j < je; j++) 
	    t += B[indx_[j] + Bi][p + Bj] * val_[j];
	  C[i + Ci][p + Cj] = t;
	}
      }
    }
    else {  // alpha != 1.0
     for (int p = 0; p < k; p++) {
	for (int i = 0; i < M_; i++) {
	  if (beta == 0.0) 
	    t = 0.0;
	  else if (beta == 1.0)
	    t = C[i + Ci][p + Cj];
	  else /* beta != 1.0 */
	    t = beta * C[i + Ci][p + Cj];
	  
	  jb = pntrb_[i];
	  je = pntre_[i];

	  for (int j = jb; j < je; j++) 
	    t += alpha * B[indx_[j] + Bi][p + Bj] * val_[j];
	  C[i + Ci][p + Cj] = t;
	}
      } 
    }
  }





  /**
    method to support matrix-vector multiplication
   */
  public void dusmv(final int transpose, final double alpha, final double x[], final int incx, 
		    final double beta,  double y[], final int incy) {

    int jb, je;
    double t;
    
    if (alpha == 0) 
      {
	if (beta != 1.0) {
	  Dense.arrayMultScalar(y, incy, beta); 
	  return;
	}
      }
    
    if (transpose == Jblas.Trans) { // calls DCSC multiply
      // need to avoid infinite recursion
      int tempTrans = Jblas.NoTrans;
      
      // create Compressed Column object
      DCSC dcsc1 = new DCSC(M_, N_, nz_, val_, indx_, pntrb_, pntre_, prop_);
      
      // multiply
      dcsc1.dusmv(tempTrans, alpha, x, incx, beta, y, incy);
      return;
    }


    if (incy == 1 && incx == 1) {

      for (int i = 0; i < M_; i++) {
	if (beta == 0.0) 
	  t = 0.0;
	else if (beta == 1.0)
	  t = y[i];
	else /* beta != 1.0 */
	  t = beta * y[i];
	
	jb = pntrb_[i];
	je = pntre_[i];
	
	for (int j = jb; j < je; j++) 
	  t += alpha * x[indx_[j]] * val_[j];
	y[i] += t;
      }
    }
    else {  // incy != 1 || incx != 1
     
      for (int i = 0; i < M_; i++) {
	if (beta == 0.0) 
	  t = 0.0;
	else if (beta == 1.0)
	  t = y[i];
	else /* beta != 1.0 */
	  t = beta * y[i];
	      
	jb = pntrb_[i];
	je = pntre_[i];
	
	for (int j = jb; j < je; j++) 
	  t += alpha * x[indx_[j] * incx] * val_[j];
	y[i * incy] += t;
      }
    }
  }



  public void dussvLD(final int transpose, 
		      final double alpha,
		      double x[], 
		      final int incx) {
    
    
    int i, j, jb, je, index;
    double z, valtmp = 0;


    if (transpose == Jblas.Trans) { // calls DCSC
      // need to avoid infinite recursion
      int tempTrans = Jblas.NoTrans;
      
      // create Compressed Column object
      DCSC dcsc1 = new DCSC(M_, N_, nz_, val_, indx_, pntrb_, pntre_, prop_);
      
      // triangular solve
      dcsc1.dussvUD(tempTrans, alpha, x, incx);

      return;
    }


    if (alpha == 0) {
      for (i = 0; i < N_ * incx; i+= incx)
	x[i] = 0;
      return;
    }

    for (i = 0; i < M_ ; i++) {

      z = 0;
    
      jb = pntrb_[i];
      je = pntre_[i];

      for(j = jb; j < je; j++) {
	index = indx_[j];
	
	if (index == i) {
	  valtmp = val_[j];
	} else {
	  z += x[index * incx] * val_[j];
	}
      }
      x[i * incx] = (x[i * incx] - z) / valtmp;
    }
    
    if (alpha != 1.0)
      for (i = 0; i < (M_ * incx); i+=incx) 
	x[i] *= alpha;
    
  }


  public void dussvLU(final int transpose, 
			final double alpha,
			double x[], 
			final int incx) {
    
    
    int i, j, jb, je, index;
    double z, valtmp = 0;


    if (transpose == Jblas.Trans) { // calls DCSC
      // need to avoid infinite recursion
      int tempTrans = Jblas.NoTrans;
      
      // create Compressed Column object
      DCSC dcsc1 = new DCSC(M_, N_, nz_, val_, indx_, pntrb_, pntre_, prop_);
      
      // triangular solve
      dcsc1.dussvUU(tempTrans, alpha, x, incx);

      return;
    }
    
    if (alpha == 0) {
      for (i = 0; i < N_ * incx; i+= incx)
	x[i] = 0;
      return;
    }

    for (i = 0; i < M_ ; i++) {
      z = 0;
      jb = pntrb_[i];
      je = pntre_[i];
      for(j = jb; j < je; j++) {
	  z += x[indx_[j] * incx] * val_[j];
      }
      x[i * incx] -= z;
    }
    
    if (alpha != 1.0)
      for (i = 0; i < (M_ * incx); i+=incx) 
	x[i] *= alpha;
    
  }



  public void dussvUD(final int transpose, 
			final double alpha,
			double x[], 
			final int incx) {
    
    
    int i, j, jb, je, index;
    double z, valtmp = 0;


    if (transpose == Jblas.Trans) { // calls DCSC
      // need to avoid infinite recursion
      int tempTrans = Jblas.NoTrans;
      
      // create Compressed Column object
      DCSC dcsc1 = new DCSC(M_, N_, nz_, val_, indx_, pntrb_, pntre_, prop_);
      
      // triangular solve
      dcsc1.dussvLD(tempTrans, alpha, x, incx);

      return;
    }
    
    if (alpha == 0) {
      for (i = 0; i < N_ * incx; i+= incx)
	x[i] = 0;
      return;
    }



    for (i=M_-1;i!=-1; i--) {
      z = 0;
      jb = pntrb_[i];
      je =  pntre_[i];
      for (j=jb+1; j<je; j++) {
	z += x[(indx_[j]) * incx] * val_[j];
      }
      x[i * incx] = (x[i * incx] - z) / val_[jb];
    }
    
    if (alpha != 1.0)
      for (i = 0; i < (M_ * incx); i+=incx) 
	x[i] *= alpha;
    
  }




  public void dussvUU(final int transpose, 
			final double alpha,
			double x[], 
			final int incx) {
    
    
    int i, j, jb, je, index;
    double z, valtmp = 0;


    if (transpose == Jblas.Trans) { // calls DCSC
      // need to avoid infinite recursion
      int tempTrans = Jblas.NoTrans;
      
      // create Compressed Column object
      DCSC dcsc1 = new DCSC(M_, N_, nz_, val_, indx_, pntrb_, pntre_, prop_);
      
      // triangular solve
      dcsc1.dussvLU(tempTrans, alpha, x, incx);

      return;
    }
    
    if (alpha == 0) {
      for (i = 0; i < N_ * incx; i+= incx)
	x[i] = 0;
      return;
    }


    for (i=M_-1;i!=-1; i--) {
      z = 0;
      jb = pntrb_[i];
      je =  pntre_[i];
      for (j=jb; j<je; j++) {
	z += x[(indx_[j])* incx] * val_[j];
      }
      x[i * incx] -= z;
    }

    if (alpha != 1.0)
      for (i = 0; i < (M_ * incx); i+=incx) 
	x[i] *= alpha;
    
  }



  /** 
    test driver - more detailed testing done by TestMult.java
   */
  public static void main(String args[]) {
    /*
      // first test main, test multiplication

    int M = 5;
    int N = 5;
    int nz = 5;
    double val[] = {1.0,1.0,1.0,1.0,1.0};
    int indx[] = {0,1,2,3,4};
    int pntrb[] = {0,1,2,3,4};
    int pntre[] = {1,2,3,4,5};
    int properties = 0;

    int transpose = Jblas.NoTrans;
    int k = 5;
    double alpha = 0.0;
    double beta = -1.0;
    //  double alpha = 2.25;
    //double beta = -3.75;

    DCSR dc1 = new DCSR(M, N, nz, val, indx, pntrb, pntre, properties); 

    dc1.print();

    double[][] B = {{2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0}};

    try {
      Dense.save(B, "out");
      B = Dense.read("out");
    }
    catch(Exception e) {
      e.printStackTrace();
    }

    Dense.print(B);

    double[][] C = new double[5][5];
    Dense.arraySetToValue(C,1.0);
        
    dc1.dusmm(transpose, k, alpha, B, beta, C);

    Dense.print(C);
    */

    /*
    // second test main - test conversion from DCOO to DCSR
    int M = 5;
    int N = 5;
    int nz = 5;
    double val[] = {1.0,1.0,1.0,1.0,1.0};
    int indx[] = {0,1,2,3,4};
    int jndx[] = {0,1,2,3,4};
    int properties = 0;

    int transpose = Jblas.NoTrans;
    int k = 5;


    DCOO dcoo1 = new DCOO(M, N, nz, val, indx, jndx, properties); 
    dcoo1.print();

    System.out.println("\n\n");

    DCSR dcsr1 = new DCSR(dcoo1);

    dcsr1.print();
    */


    // 3rd test main, test conversion again    
    int M;
    int N;
    int nz;
    
    DCOO dcoo1 = new DCOO();
    DCSR dcsr1;

    try {
      dcoo1.read("Data/pores_1.mtx");
    }
    catch (IOException e) {
      e.printStackTrace();
    }
     
    M = dcoo1.getNumRows();
    N = dcoo1.getNumCols();
    nz = dcoo1.getNumNonZero();

    // convert the DCOO object A into the DCSR object CSRA
    dcsr1 = new DCSR(dcoo1);

    System.out.println("DCOO A:");
    
    dcoo1.print();

    System.out.println("\nDCSR CSRA:");

    dcsr1.print2();
  }
}
