/**

  DCSC.java - double-precision Compressed Column Format

  @author Galen J. Wilkerson

 */

import java.io.*;

public class DCSC {
  
  int M_, N_, nz_;
  int prop_;
  double[] val_;          /* length = nz */
  int[] indx_;             
  int[] pntrb_;             
  int[] pntre_;             


  /** 
    constructor
   */
  public DCSC( final int M, final int N, final int nz, final double[] val, final int[] indx, 
	       final int[] pntrb, final int[] pntre, final int properties) 
  {
    M_ = M;
    N_ = N;
    nz_ = nz;

    val_ = val;
    indx_ = indx;
    pntrb_ = pntrb;
    pntre_ = pntre;

    prop_ = properties;
  }


  /**
    conversion constructor - converts from DCOO -> DCSC
   */
  public DCSC (DCOO A) {
    int tempi=0;
    int tempM = A.M_;
    int tempN = A.N_;
    int tempI[] = A.indx_;
    int tempJ[] = A.jndx_;
    double tempval[] = A.val_;
    int tempnz = A.nz_;
    int tempproperties = A.prop_;
    int tempindex = 0;
   
    /* components for CSC routine */
    double csc_val[] = new double[tempnz];
    int csc_pntrb[] = new int[tempN];
    int csc_pntre[] = new int[tempN];
    int csc_j[] = new int[tempnz];
    int tally[] = new int[tempN];
    
    /* 1st pass: find out how many elements in each row */

    for (int i=0; i<tempnz; i++)
      tally[tempJ[i]]++;

    /* convert tally[] into (0-based) row pointer by accumulating
       row-lengths.  tally[i] now points to the begining or row i  */
    
    for (int i = 1;i < tempN; i++)
      tally[i] += tally[i-1];

    /* copy these values into pntrb[]; (we will be using the values
       of tally[] as place-markers, which change as we fill up the
       CSC matrix). */

    csc_pntrb[0]=0;
    for (int i=0; i < tempN-1; i++) {
      csc_pntrb[i+1] = tally[i];
      csc_pntre[i] = tally[i];
    }
    csc_pntre[tempN-1] = tally[tempN-1];

    /* copy back original row-starts into tally */
    for (int i=0; i < tempN; i++)
      tally[i] = csc_pntrb[i];

    /* 2nd pass: now go through and fill-in CSC format */
    for (int i=0; i < tempnz; i++)
      {
	tempindex = tally[tempJ[i]]++;
	csc_val[tempindex] = tempval[i];
	csc_j[tempindex] = tempI[i];
      }

    //    return create_DCSC(M, N, nz, csc_val, csc_j, csc_pntrb, csc_pntrb+1, 
    //	       properties);

    M_ = tempM;
    N_ = tempN;
    nz_ = tempnz;

    val_ = csc_val;
    indx_ = csc_j;
    pntrb_ = csc_pntrb;
    pntre_ = csc_pntre;

    prop_ = tempproperties;
    return;
    
  }



  public int getNumRows() {
    return M_;
  }

  public int getNumCols() {
    return N_;
  }

  public int getNumNonZero() {
    return nz_;
  }

  /**
    print the current object
   */
  public void print() {
    System.out.println("M " + M_);
    System.out.println("N " + N_);
    System.out.println("nz " + nz_);

    System.out.println("val\n");
    for (int i = 0; i < val_.length; i++) {
      System.out.println(val_[i] + " ");
    }

    System.out.println("indx\n");
    for (int i = 0; i < val_.length; i++) {
      System.out.println(indx_[i] + " ");
    }

    System.out.println("pntrb\n");
    for (int i = 0; i < pntrb_.length; i++) {
      System.out.println(pntrb_[i] + " ");
    }

    System.out.println("pntre\n");
    for (int i = 0; i < pntre_.length; i++) {
      System.out.println(pntre_[i] + " ");
    }

    System.out.println("properties " + prop_);
  }

  /**
    method to support CaABbC matrix-matrix multiplication.
   */
  public void dusmm(final int transpose, final int k, final double alpha,
			  final double B[][], final double beta, double C[][]) 
  { 

    double t;
    int jb, je;

    if (alpha == 0) 
      {
	if (beta != 1.0) {
	  Dense.arrayMultScalar(C, beta); 
	  return;
	}
      }
    
    // handle transpose with DCSR
    if (transpose == Jblas.Trans) { 
    
      // need to avoid infinite recursion
      int tempTrans = Jblas.NoTrans;
      
      // create Compressed Row object
      DCSR dcsr1 = new DCSR(M_, N_, nz_, val_, indx_, pntrb_, pntre_, prop_);
      
      // multiply
      dcsr1.dusmm(tempTrans, k, alpha, B, beta, C);

      return;
    }

    
    if (beta == 0.0) 
      Dense.arraySetToValue(C,0.0);
    else if (beta != 1.0) 
      Dense.arrayMultScalar(C,beta);

    if (alpha == 1.0) {
      for (int p = 0; p < k; p++) {
	for (int i = 0; i < N_; i++) {
	  
	  jb = pntrb_[i];
	  je = pntre_[i];
	  
	  for (int j = jb; j < je; j++) 
	    C[indx_[j]][p] += B[i][p] * val_[j];
	}
      }
    }
    else {  // alpha != 1.0
      for (int p = 0; p < k; p++) {
	for (int i = 0; i < N_; i++) {
	  
	  jb = pntrb_[i];
	  je = pntre_[i];
	  
	  for (int j = jb; j < je; j++) 
	    C[indx_[j]][p] += alpha * B[i][p] * val_[j];
	}
      }
    }
  }


  /**
    method to support CaABbC matrix-matrix multiplication using submatrices
    */
  public void dusmm(final int transpose, final int k, final double alpha, final double B[][], 
		    final int Bi, final int Bj, final double beta, double C[][], final int Ci, final int Cj)
  { 

    double t;
    int jb, je;


    if (alpha == 0) 
      {
	if (beta != 1.0) {
	  Dense.arrayMultScalar(C, beta); 
	  return;
	}
      }

    // handle this with DCSR 
    if (transpose == Jblas.Trans) {
    
      // need to avoid infinite recursion
      int tempTrans = Jblas.NoTrans;
      
      // create Compressed Row object
      DCSR dcsr1 = new DCSR(M_, N_, nz_, val_, indx_, pntrb_, pntre_, prop_);
      
      // multiply
      dcsr1.dusmm(tempTrans, k, alpha, B, Bi, Bj, beta, C, Ci, Cj);
      
      return;
  
    }
    

    if (beta == 0.0) 
      Dense.arraySetToValue(C,0.0);
    else if (beta != 1.0) 
      Dense.arrayMultScalar(C,beta);


    if (alpha == 1.0) {
      for (int p = 0; p < k; p++) {
	for (int i = 0; i < N_; i++) {
	  
	  jb = pntrb_[i];
	  je = pntre_[i];
	  
	  for (int j = jb; j < je; j++) 
	    C[indx_[j] + Ci][p + Cj] += B[i + Bi][p + Bj] * val_[j];
	}
      }
    }
    else {  // alpha != 1.0
      for (int p = 0; p < k; p++) {
	for (int i = 0; i < N_; i++) {

	  jb = pntrb_[i];
	  je = pntre_[i];
	  
	  for (int j = jb; j < je; j++) 
	    C[indx_[j] + Ci][p + Cj] += alpha * B[i + Bi][p + Bj] * val_[j];
	}
      }
    }
  }


  /** 
    test driver - more detailed testing done by TestMult.java
   */
  public static void main(String args[]) {
    int M = 5;
    int N = 5;
    int nz = 5;
    double val[] = {1.0,1.0,1.0,1.0,1.0};
    int indx[] = {0,1,2,3,4};
    int pntrb[] = {0,1,2,3,4};
    int pntre[] = {1,2,3,4,5};
    int properties = 0;

    int transpose = Jblas.NoTrans;
    int k = 5;
    double alpha = 1.0;
    double beta = 0.0;
    //  double alpha = 2.25;
    //double beta = -3.75;

    DCSC dc1 = new DCSC(M, N, nz, val, indx, pntrb, pntre, properties); 

    dc1.print();

    double[][] B = {{2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0}};

    Dense.print(B);

    double[][] C = new double[5][5];
    Dense.arraySetToValue(C,1.0);
        
    //    dc1.dusmm(transpose, k, alpha, B, beta, C);
    dc1.dusmm(transpose, k, alpha, B,0,0, beta, C,0,0);

    Dense.print(C);
  
  }
}
