/**

  DCOO.java - double-precision Compressed Coordinate format 

  @author Galen J. Wilkerson

 */

import java.io.*;

public class DCOO {
  
  int M_, N_, nz_;
  int prop_;
  double[] val_;          /* length = nz */
  int[] indx_;             
  int[] jndx_;             


  public DCOO() {
  }


  /** 
    constructor
   */
  public DCOO( final int M, final int N, final int nz, final double[] val, final int[] indx, 
	       final int[] jndx, final int properties) 
  {
    M_ = M;
    N_ = N;
    nz_ = nz;

    /*
    val_ = new double[val.length];
    indx_ = new int[indx.length];
    jndx_ = new int[jndx.length];
    
    for (int i = 0; i < val.length; i++) {
      val_[i] = val[i];
      indx_[i] = indx[i];
      jndx_[i] = jndx[i];
    }
    */

    val_ = val;
    indx_ = indx;
    jndx_ = jndx;

    prop_ = properties;
  }


  public int getNumRows() {
    return M_;
  }

  public int getNumCols() {
    return N_;
  }

  public int getNumNonZero() {
    return nz_;
  }

  /**
    print the current object
   */
  public void print() {
    System.out.println("M " + M_);
    System.out.println("N " + N_);
    System.out.println("nz " + nz_);

    System.out.println("val   indx    jndx");
    for (int i = 0; i < val_.length; i++) {
      System.out.println(val_[i] + "       " + indx_[i] + "       " + jndx_[i]);
    }

    System.out.println("properties " + prop_);
  }


  /** 
    Read the current object from Matrix-Market format.
   */
  public void read(final String fileName) throws IOException {
    NumericReader nr = new NumericReader(new FileReader(fileName));

    nr.readLine();  // read the header

    int height = nr.readInt();
    int width = nr.readInt();
    int nz1 = nr.readInt();
    
    // read the sparse array
    int[] tempIndx = new int[nz1];
    int[] tempJndx = new int[nz1];
    double[] tempVal = new double[nz1];

    for (int i = 0; i < nz1; i++) {
      tempIndx[i] = nr.readInt() - 1;  // convert to zero-base
      tempJndx[i] = nr.readInt() - 1;
      tempVal[i] = nr.readDouble();
    }
    
    
    M_ = height;
    N_ = width;
    nz_ = nz1;
    val_ = tempVal;
    indx_ = tempIndx;
    jndx_ = tempJndx;
    
    return;
  }



  /**
    Save the current object in Matrix-Market format.
   */
  public void saveSparse(String fileName) throws IOException {
  
    PrintWriter pw = new PrintWriter(new FileWriter(fileName));

    pw.println("%%MatrixMarket matrix coordinate real general");
    pw.println(M_ + " " + N_ + " " + nz_);

    for (int i = 0; i < nz_; i++) {
      pw.println((indx_[i] + 1) + " " + (jndx_[i] + 1) + " " + val_[i]);
    }

    pw.close();

  }



  /**
    method to support CaABbC matrix-matrix multiplication.
   */
  public void dusmm(final int transpose, final int k, final double alpha,
			  final double B[][], final double beta, double C[][]) 
  { 

    int[] tempIndx = indx_;
    int[] tempJndx = jndx_;

    if (alpha == 0) 
      {
	if (beta != 1.0) {
	  Dense.arrayMultScalar(C, beta); 
	  return;
	}
      }
    
    if (transpose == Jblas.Trans) { // swap indx_ and jndx_ 

      // int[] temp = indx_;
      // indx_ = jndx_;
      // jndx_ = temp;
      
      tempIndx = jndx_;
      tempJndx = indx_;
    }

    if (beta == 0.0) 
      Dense.arraySetToValue(C,0.0);
    else if (beta != 1.0) 
      Dense.arrayMultScalar(C,beta);

    if (alpha == 1.0) {
      for (int i = 0; i < k; i++) 
	for (int j = 0; j < nz_; j++) {
	  C[tempIndx[j]][i] += val_[j] * B[tempJndx[j]][i];
	}
    }
    else {  // alpha != 1.0
      for (int i = 0; i < k; i++) 
	for (int j = 0; j < nz_; j++) {
	  C[tempIndx[j]][i] += val_[j] * B[tempJndx[j]][i] * alpha;
	}
    }
  }




  /**
    method to support CaABbC matrix-matrix multiplication using submatrices
    */

  public void dusmm(final int transpose, final int k, final double alpha, final double B[][], 
			  final int Bi, final int Bj, final double beta, double C[][], final int Ci, final int Cj)
  { 
    int[] tempIndx = indx_;
    int[] tempJndx = jndx_;
    
    if (alpha == 0) 
      {
	if (beta != 1.0) {
	  Dense.arrayMultScalar(C, beta); 
	  return;
	}
      }
    
    if (transpose == Jblas.Trans) { // swap indx_ and jndx_ 
      // int[] temp = indx_;
      // indx_ = jndx_;
      // jndx_ = temp;
      
      tempIndx = jndx_;
      tempJndx = indx_;
    }

    if (beta == 0.0) 
      Dense.arraySetToValue(C,0.0);
    else if (beta != 1.0) 
      Dense.arrayMultScalar(C,beta);

    if (alpha == 1.0) {
      for (int i = 0; i < k; i++) 
	for (int j = 0; j < nz_; j++) {
	  C[(tempIndx[j] + Ci)][(i + Cj)] += val_[j] * B[(tempJndx[j] + Bi)][i + Bj];
	}
    }
    else {  // alpha != 1.0
      for (int i = 0; i < k; i++) 
	for (int j = 0; j < nz_; j++) {
	  C[tempIndx[j] + Ci][(i+ Cj)] += val_[j] * B[(tempJndx[j] + Bi)][i + Bj] * alpha;
	}
    }
  }



  /** 
    test driver - more detailed testing done by TestMult.java
   */
  public static void main(String args[]) {
    int M = 5;
    int N = 5;
    int nz = 5;
    double val[] = {1.0,1.0,1.0,1.0,1.0};
    int indx[] = {0,1,2,3,4};
    int jndx[] = {0,1,2,3,4};
    int properties = 0;

    int transpose = Jblas.NoTrans;
    int k = 5;
    double alpha = 2.25;
    double beta = -3.75;

    DCOO dc1 = new DCOO(M, N, nz, val, indx, jndx, properties); 

    dc1.print();

    double[][] B = {{2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0},
		    {2.0,2.0,2.0,2.0,2.0}};

    try {
      Dense.save(B, "out");
      B = Dense.read("out");
    }
    catch(Exception e) {
      e.printStackTrace();
    }

    Dense.print(B);

    double[][] C = new double[5][5];
    Dense.arraySetToValue(C,1.0);
        
    dc1.dusmm(transpose, k, alpha, B, beta, C);

    Dense.print(C);
  
    try {
      dc1.read("Data/fidap005.mtx");
    }
    catch(Exception e) {
      e.printStackTrace();
    }

    dc1.print();
  }
}
